/*
 *	Name: MuxParm.h
 *	Description: Global point for setting up the parameters used by the Mux
 *	Project.
 *	Author: Mark Salyzyn January 15 1994
 */

/*
 *	NINSTANCE
 *
 *	This parameter is set to the maximum number of instances to support.
 *	I could have done this dynamically, but chose a static size for
 *	simplicity for now.
 */
#define NINSTANCE	8		/* Number of Instances to support */

/*
 *	NTTY
 *
 *	This parameter is used to set the maximum number of ports to support
 *	for each instance of the Mux driver. All the ports in each instance
 *	must be tied to the same interrupt, and need not have a multiplex
 *	port available to streamline the driver's overhead. NTTY should be set
 *	to the number of ports available on the largest multiplexed serial
 *	board. The number of ports that can be supported, then, is
 *	NINSTANCE * NTTY.
 */

#define NTTY	4			/* Number of Serial Ports/Instance */

/*
 *	NO_MUX
 *
 *	Define this if you desire to reduce the interrupt overhead on your
 * system if you have no multiplex register on a shared interrupt device, or
 * if you only have single ports on your system for each interrupt. If you
 * define NTTY as 1, then NO_MUX is automatically defined.
 */
#define NO_MUX	1			/* No multiplexor port		*/

/*
 * Input Buffer.
 *
 * It is prefered to make the buffers programmable, via "Buffer Size [#]",
 * "RTS Low Water [#]" and "RTS High Water [#]" variables in the
 * Instance?.table. The default buffer size will be DOUBLE_BUFFER, the default
 * low water mark will be one third the programmed buffer size, and the
 * default high water mark will be half way between the low water mark and the
 * buffer size.
 */

#define DOUBLE_BUFFER	8192


/*
 * The Receive fifo interrupt trigger point. If it is the intention to use
 * very high baud rates, the default trigger should be set lower. It has been
 * found on some systems that setting FIFO_TRIGGER to FIFO_TRIGGER_1 may solve
 * losing characters on streaming protocols at 38400 Baud on machines that
 * appear burdened with interrupt latency on other hardware.
 */
#define FIFO_TRIGGER   FIFO_TRIGGER_4

/*
 * Name used by Project.
 */

static char MuxDeviceName[] = "Mux";

static char MuxDeviceKind[] = "Serial";	/* Kind of driver registered	*/
