/*
 *	Name: MuxParm.h
 *	Description: Global point for setting up the parameters used by the Mux
 *	Project.
 *	Author: Mark Salyzyn January 15 1994
 */

#define NINSTANCE	4		/* Number of Instances to support */

#define NTTY	8			/* Number of Serial Ports/Instance */

/*
 * Input Buffer.
 *
 * If the value is positive, we allocate a static buffer for every single port,
 * regardless if the port is open or not (you may wish to adjust the NINSTANCE
 * and NTTY to reduce the memory requirements). This offers good receive
 * performance, and will always function regardless of the state of the
 * kernel memory supply.
 *
 * If it is prefered to make the buffers programmable, via "Buffer Size [#]",
 * "RTS Low Water [#]" and "RTS High Water [#]" variables in the
 * Instance?.table, then make this value negative. The default buffer size will
 * be the negative of DOUBLE_BUFFER, the default low water mark will be one
 * third the programmed buffer size, and the default high water mark will be
 * half way between the low water mark and the buffer size.
 *
 * If the value is left unspecified, then we have no buffering, and lose RTS
 * handshaking. This reduces the processor load by a small margin, due to
 * serial overhead, but will not handle 38400 Baud (on a 486DX2/66) reliably
 * and will lose characters at that rate.
 */
#define DOUBLE_BUFFER	-2048

/*
 * Name used by Project. If you set this to "SerialPorts", the driver should
 * then be close to a drop in replacement for the existing NeXTSTEP serial
 * driver (Minus the fact that we set up a separate instance for each serial
 * port installed in the system). This should get around any hard coded
 * routines that communicate to the serial port driver assuming that the only
 * serial communications driver is the one supplied by NeXT. I have not tested
 * at this moment if the change in name will allow the SerialMouse to work
 * with my driver since there is a published workaround by mixing the supplied
 * SerialPorts driver and this driver.
 */
static char MuxDeviceName[] = "Mux";

static char MuxDeviceKind[] = "Serial";	/* Kind of driver registered	*/
