/*
 *	Name: MuxParm.h
 *	Description: Global point for setting up the parameters used by the Mux
 *	Project.
 *	Author: Mark Salyzyn January 15 1994
 */
/*
 *	NINSTANCE
 *
 *	This parameter is set to the maximum number of instances to support.
 *	I could have done this dynamically, but chose a static size for
 *	simplicity for now.
 */
#define NINSTANCE	4		/* Number of Instances to support */

/*
 *	NTTY
 *
 *	This parameter is used to set the maximum number of ports to support
 *	for each instance of the Mux driver. All the ports in each instance
 *	must be tied to the same interrupt, and need not have a multiplex
 *	port available to streamline the driver's overhead. NTTY should be set
 *	to the number of ports available on the largest multiplexed serial
 *	board. The number of ports that can be supported, then, is
 *	NINSTANCE * NTTY.
 */

#define NTTY	8			/* Number of Serial Ports/Instance */

/*
 *	NO_MUX
 *
 *	Define this if you desire to reduce the interrupt overhead on your
 * system if you have no multiplex register on a shared interrupt device, or
 * if you only have single ports on your system for each interrupt. If you
 * define NTTY as 1, then NO_MUX is automatically defined.
 */
/* #define NO_MUX	1		/* No multiplexor port		*/

/*
 * Input Buffer.
 *
 * It is prefered to make the buffers programmable, via "Buffer Size [#]",
 * "RTS Low Water [#]" and "RTS High Water [#]" variables in the
 * Instance?.table. The default buffer size will be DOUBLE_BUFFER, the default
 * low water mark will be one third the programmed buffer size, and the
 * default high water mark will be half way between the low water mark and the
 * buffer size.
 */
#define DOUBLE_BUFFER	3328

/*
 * The Receive fifo interrupt trigger point. If it is the intention to use
 * very high baud rates, the default trigger should be set lower. It has been
 * found on some systems that setting FIFO_TRIGGER to FIFO_TRIGGER_1 may solve
 * losing characters on streaming protocols at 38400 Baud on machines that
 * appear burdened with interrupt latency on other hardware.
 */
#define FIFO_TRIGGER   FIFO_TRIGGER_4

/*
 *	Priorities of various levels of the Mux serial Driver, higher
 * priorities may be necessary to improve performance in certain
 * configurations. Although Driver Design says we should be at IPLTTY, this
 * was in the days of PDP11s, in a 386 system, IPLTTY = SOFTINT0 = 1 which
 * makes things undesireable. Do not change this variable unless you know what
 * you are doing (it is currently untested at any other value).
 */
#define MUX_IPL	1

/*
 *	_NX_KERNEL
 *
 *	This parameter may be used later in the code adjusting for various
 *	differences in the Kernel. Currently, only 320 is supported, as 310
 *	does not work with Mux, and 330 problems have not been entirely
 *	checked out. Do not change this variable unless you know what you are
 *	doing.
 */
#define _NX_KERNEL	320

/*
 *	POSIX_KERN
 *
 *	This parameter is used to add the TIOCSPGRP and TIOCGPGRP ioctl call
 *	enhancements. Also, fixes associated with kernel structures are
 *	included. Do not undefine this variable unless you know what you are
 *	doing.
 */
#define POSIX_KERN	1

/*
 *	MuxInspector
 *
 *	This parameter compiles the very borken and never finished Mux
 *	Device Inspector. Define this if you intend to finish the project.
 */
#undef MuxInspector

/*
 * Name used by Project.
 */

static char MuxDeviceName[] = "Mux";

static char MuxDeviceKind[] = "Serial";	/* Kind of driver registered	*/
