/*
 * Copyright (c) 1991 The Regents of the University of California.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	@(#)com.c		7.5 (Berkeley) 5/16/91
 *	@(#)Mux8.m		1.0 (Salyzyn)	93/11/16
 *	@(#)Mux.m		1.1 (Salyzyn)	94/01/13
 *		- Fixed the handling of Parity once again, I hope I got it
 *		  right this time for the BSD universe. Added CREAD, PARENB
 *		- Cleaned up the locking code between the cu? and ttyd? ports.
 *		- Root *must* be able to override XCLUSIV locking, I don't
 *		  like this, but getty re-opens the port ... I *could* allow
 *		  the same process group to reopen instead but this would
 *		  then be non-standard operation (but may be required in end).
 *		- Added preliminary AST support, still no board to test it
 *		  with ...
 *	@(#)Mux.m		1.2 (Salyzyn) 94/01/23
 *		- Added Double Buffered operation to ensure reliable 38400
 *		  reception of data. Performance improves if you disable this
 *		  feature if you intend to operate at 19200 for some protocols.
 *		  With the addition of this support, I could add sane RTS side
 *		  of th RTS/CTS handshaking. RTS handshaking is done *always*
 *		  which *could* cause problems for some hardware arrangements.
 *		  I thank Markus Wenzel <next2@info2.rus.uni-stuttgart.de> for
 *		  helping me test this code on his 16800 Baud Modem setup with
 *		  Taylor UUCP.
 *		- Once again I know nothing about locking, and should use
 *		  the NXLock and NXConditionLock methods to work out the
 *		  details. I have expanded the semaphores and `manual' locking
 *		  and improvements are noticed. I still get a *low throughput*
 *		  mode of operation once in a while with bi-directional lines
 *		  but no loss of characters (?). I am beginning to suspect that
 *		  NeXTSTEP has a maximum throughput that it can hack, and I am
 *		  observing the results of the OS trying to deal with it.
 *		- tidied up some of the placement of spltty code to improve the
 *		  latency.
 *	@(#)Mux.m		1.3 (Salyzyn) 94/02/05
 *		- Added dynamically (Software controllable sizes in the
 *		  Instance0 tables) allocated receive buffers, and programmable
 *		  low and high water marks for RTS handling.
 *		- recognized AST board handling under configuration control,
 *		  rather than by a manifest compile constant. To program for
 *		  the AST board, the mux takes up a port range of 1. To program
 *		  for a DigiBoard 8, the mux range is specified as a range of
 *		  2. To program for wired-or (dumb) interrupt cards, do not
 *		  specify a mux port.
 *		- removed more processing code from interrupt and moved it into
 *		  the non-interrupt Executive task to improve the latency
 *		  figures even more.
 *		- One more time, I popped into the locking code. This time
 *		  a subtle change where instead of letting root re-open the
 *		  port *any* time it wants, I will *only* let the same
 *		  *mach* task re-open. This should solve the problem with
 *		  regular (non super) users running other communications
 *		  programs.
 *	@(#)Mux.m		1.4 (Salyzyn) 94/02/08
 *		- DCD should *only* be checked for /dev/ttyd?? entry points
 *		  as pointed out by James Cassidy <jcassidy@quark.uwaterloo.ca>
 *		- DCD should not be detected when dropped by the `LOCAL' ports
 *		  as pointed out by James E Quick <jq@localhost.phcs.com>
 *		- at 38400 and 57600, it appears that we need a little extra
 *		  head start to handle receive Interrupts. Set default
 *		  FIFO_TRIGGER to 4. I now lose 1 character every million
 *		  at 57600 Baud (from a 14400 Baud modem though, so this is
 *		  intellectually dishonost!). I could not notice any change
 *		  in CPU overhead, but would appreciate knowing what people
 *		  noticed on slower machines.
 *		- Removed the use of XCLUDE for bidirectional locking. The
 *		  use of this flag caused a great deal of trouble with TipTop,
 *		  as pointed out by Amir <amir@avalon.unizh.ch>, and with
 *		  screens (I can find bugs too eh?) and other programs. Pity ...
 *		- Added Transmit buffering to improve Taylor `i' protocol UUCP
 *		  performance with it's capability of setting larger buffer
 *		  sizes. Have managed upward of 3600ch/s at 57600 to a 14400
 *		  modem on a compressible file!
 *		- With Buffer Size not set in these original Instance?.tables
 *		  V1.3 would `crash and burn'. The problem was when the
 *		  variables were not set, the valid hiwater mark was placed
 *		  into the lowater placeholder, and hiwater still remained
 *		  invalid.
 *	@(#)Mux.m		1.5 (Salyzyn) 94/05/15
 *		- Allowed the software to compile with NO_MUX to provide a
 *		  slightly lower overhead on systems that have no interrupt
 *		  multiplexer port.
 *		- Turned off Modem state interrupts when the port is opened
 *		  in modes that ignore it.
 *		- Allowed per-port configuration of the FIFO trigger level in
 *		  the instance table. "FIFO #" = "1"|"4"|"8"|"14"
 *		- Fixed a possible bug in Muxread calling Muxintr at low
 *		  priority, rather than at SPLTTY.
 *		- There is a state the driver can enter that locks out access
 *		  to the cu? ports even though the ttyd? ports are in a WAIT,
 *		  solved this problem by adding more checking to the BUSY
 *		  locking code.
 *		- Added in code to deal with noisy modem lines (appears to
 *		  occur with a higher probability on 16450 based systems).
 *		  Thanks goes out to Christian Starkjohann
 *		  <next@merlin.par.univie.ac.at> for taking some time to
 *		  investigate this and other 16450 problems in great detail.
 *		- Changed the code so we can pick the priority level. This
 *		  is necessary because on some machine it appears picking
 *		  IPLTTY is somewhat low for *their* priorities.
 *	@(#)Mux.m		1.6 (Salyzyn) 94/12/25
 *		- Added the TIOCSPGRP and TIOCGPGRP overrides in ioctl to solve
 *		  a problem with some applications losing access to these
 *		  because of a Kernel Bug. Not sure if these are performed
 *		  correctly, as there are some problems with `/dev/tty',
 *		  (and thus with `screen' and `csh') operating because of this
 *		  fixup. These fixups are needed so that mgetty+sendfax,
 *		  NXFax and friends can function, so it is a tradeoff at this
 *		  moment until I can dig deaper into the vulgarities.
 *		- By defining POSIX_KERN, we solve a whole host of bugs that
 *		  didn't show in the 3.2 Kernel, but are readily apparent in
 *		  the 3.3 Kernel.
 *	@(#)Mux.m		1.7 (Salyzyn) 95/2/10
 *		- The modem line discipline should not be called in all cases
 *		  of modem transition, as it appears NS3.3 is *more* UNIX
 *		  compliant now! This affected NXFax operation.
 *		- Finally added the flushing of the double buffer on VINTR and
 *		  with flush on interrupt enabled.
 *		- Removed the NXFAX_BUG code, no need any longer.
 */

#import "Mux.h"

#if NTTY > 0
/*
 * COM driver, based on HP dca driver
 * uses National Semiconductor NS16450/NS16550AF UART and 82510 devices.
 *
 * Includes modifications based on those of
 * brad huntting <huntting@glarp.co> and others
 * for CTS, RTR (RTS for receive), etc. (Mark Salyzyn, HA! there was NO
 * RTS handling in Brad Huntting's code ...)
 *
 * Meant to support the DigiBoard 4/8 multiplexed card (two multiplex registers
 * depending on odd/even interrupt returns a bit indicating which port(s) are
 * triggered) or AST multiplexed card (one multiplex register, with bit as
 * for DigiBoard).
 */

#if (defined(POSIX_KERN))
	typedef int	pid_t;
#endif
#import <sys/param.h>
#import <sys/systm.h>
#import <sys/ioctl.h>
#import <sys/tty.h>
#import <kernserv/prototypes.h>	/* To handle process information */
#import <sys/proc.h>
#import <sys/user.h>
#import <sys/conf.h>
#import <sys/file.h>
#import <sys/uio.h>
#import <sys/kernel.h>
#import <sys/syslog.h>
#import <sys/errno.h>
#import <sys/callout.h>

#import <driverkit/generalFuncs.h>
#import	<driverkit/devsw.h>
#import <kernserv/printf.h>
#import	<kernserv/i386/spl.h>
#import <kernserv/lock.h>
#import <driverkit/i386/ioPorts.h>

#import "comreg.h"
#import "ns16550.h"
#import "intr.h"    /* Undocumented NeXTSTEP Mach Kernel Interrupt Handlers */

/*
 *	This fixes a bug in NXFax that sets the serial port to 5 bits/character
 *		Thanks goes out to Christian Starkjohann for finding this.
 *
 */

extern int	getc();

#if (NTTY == 1)
#	define	NO_MUX	1
#endif
#if (MUX_IPL == 7)	/* Yipes!!! */
#	define splmux()	spl7()
#elif (MUX_IPL == 6)
#	define splmux()	spl6()
#elif (MUX_IPL == 5)
#	define splmux()	spl5()
#elif (MUX_IPL == 4)
#	define splmux()	spl4()
#elif (MUX_IPL == 3)
#	define splmux()	spl3()
#elif (MUX_IPL == 2)
#	define splmux()	spl2()
#elif (MUX_IPL == 1)
#	define splmux()	spl1()
#endif

/*
 *	Thanks comes from Philippe Prindeville for pointing out how to *realy*
 * get ahold of the process structure.
 */
#undef u
#undef current_thread

static inline thread_t
current_thread(void)
{	extern thread_t active_threads[];

	return active_threads[0];
}
extern struct proc *proc_from_thread(thread_t);
extern struct uthread *uthread_from_thread(thread_t);
extern struct utask *utask_from_thread(thread_t);

#define getproc()	(proc_from_thread(current_thread()))
/* Per/thread `u' area */
#define u_thread()	(uthread_from_thread(current_thread()))
#define u_task()	(utask_from_thread(current_thread()))

struct tty_sc	{
	int		sc_flags;	/* definitions below		*/
	unsigned char	sc_mcount;	/* number of modem transitions	*/
	int		sc_mcr;		/* MCR_IENABLE for single ports	*/
					/* MCR_INTROFF for AST mux	*/
	int		sc_addr;	/* IO base address		*/
#	ifndef NO_MUX
		int	sc_mux;		/* IO multiplex address		*/
		int	sc_bit;		/* Bit map of multiplex		*/
#	endif
	task_t		sc_task;	/* The current task for locking	*/
	struct tty	sc_tty;		/* Our copy of the ioctl bits	*/
	struct nty *	sc_nty;		/* nty termios structure	*/

	char *		sc_kalloc;	/* The *real* wired down buffer	*/
	int		sc_ksize;	/* The allocated size of the buffer */
	unsigned	sc_size;	/* The requested size of buffer	*/

	int		sc_rts_lowat;	/* The low water mark for RTS	*/
	int		sc_rts_hiwat;	/* The high water mark for RTS	*/

	short *		sc_rxtop;	/* The top of the allocated buffer */
	short *		sc_rxbuf;	/* Allocated receive Buffer	*/

	char *		sc_txtop;	/* The top of the allocated buffer */
	char *		sc_txbuf;	/* Allocated transmit Buffer	*/

	int		sc_rxnum;	/* Buffer size maintenance	*/
	short *		sc_rxin;	/* Input from Interrupts	*/
	short *		sc_rxout;	/* Output to CLIST		*/
	char *		sc_txin;	/* Input from kernel		*/
	char *		sc_txout;	/* Output in Interrupts		*/
};

#define RTS_LOWAT(s)	((s)->sc_rts_lowat)
#define RTS_HIWAT(s)	((s)->sc_rts_hiwat)
#define TOP_RXBUF(s)	((s)->sc_rxtop)
#define TOP_TXBUF(s)	((s)->sc_txtop)

static struct tty_sc sc[NINSTANCE * NTTY];

/*
 *	The sc_rxnum variable does not need to be accurate, it will be
 *	resynchronized once in a while through the normal clearing that
 *	occurs of the buffer pointers. It is used to handle the RTS handshaking
 *	more efficiently than if we calculated the space left in the buffer.
 *
 *	MCR handles the RTS signalling. This is done as a macro dubiously to
 *	handle this quickly.
 */
#define MCR(s)	((((s)->sc_size) \
	 ? (((s)->sc_rxnum > RTS_HIWAT(s)) \
	  ? ((s)->sc_tty.t_state |= TS_INPUTFULL) \
	  : ((s)->sc_rxnum < RTS_LOWAT(s)) \
	   ? ((s)->sc_tty.t_state &= ~TS_INPUTFULL) \
	   : 0) \
	 : ((((s)->sc_tty.t_rawq.c_cc + (s)->sc_tty.t_canq.c_cc) \
		> TTHIWAT (&(s)->sc_tty)) \
	  ? ((s)->sc_tty.t_state |= TS_INPUTFULL) \
	  : (((s)->sc_tty.t_rawq.c_cc + (s)->sc_tty.t_canq.c_cc) \
		< TTLOWAT (&(s)->sc_tty)) \
	   ? ((s)->sc_tty.t_state &= ~TS_INPUTFULL) \
	   : 0)), \
	(((s)->sc_tty.t_state & TS_INPUTFULL) \
	    ? ((s)->sc_mcr &= ~MCR_RTS) : ((s)->sc_mcr |= MCR_RTS)))

#if (NINSTANCE*NTTY) > 32
# error	"Jane, you ignorant slut. Don't you know your cut off at 32?"
#endif

/*
 * I am using shared approach here, since Object variables won't give me the
 * information I need elsewhere.
 */
static int	num_devices[NINSTANCE];	/* number of ports per each instance */
static int	num_instance;		/* Number of instances registered */

/* First entry lookup table into the sc structure for the specified instance */
static struct tty_sc *	first_sc[NINSTANCE];

/* sc_flags */
#define	SC_HASFIFO	0x0001	/* is 16550 with fifo (use default value) */
#define SC_HASFIFO_1	0x0000	/* Use 1B FIFO Trigger (best performance) */
#define SC_HASFIFO_4	0x0002	/* Use 4B FIFO Trigger */
#define SC_HASFIFO_8	0x0004	/* Use 4B FIFO Trigger */
#define SC_HASFIFO_14	0x0006	/* Use 14B FIFO Trigger (lowest overhead) */

#define SC_BORKENFIFO	0x0008	/* is a 16550 with a broken fifo */
#define SC_HAS4FIFO	0x0010	/* is an 82510 with four byte Fifo */
#define SC_HASCTS	0x8000	/* has the CTS signal available in hardware */

#ifndef	FIFO_TRIGGER
#	define FIFO_TRIGGER		FIFO_TRIGGER_4
#endif
#if (FIFO_TRIGGER) == (FIFO_TRIGGER_1)
#	define SC_HASFIFO_DEFAULT	SC_HASFIFO_1
#endif
#if (FIFO_TRIGGER) == (FIFO_TRIGGER_4)
#	define SC_HASFIFO_DEFAULT	SC_HASFIFO_4
#endif
#if (FIFO_TRIGGER) == (FIFO_TRIGGER_8)
#	define SC_HASFIFO_DEFAULT	SC_HASFIFO_8
#endif
#if (FIFO_TRIGGER) == (FIFO_TRIGGER_14)
#	define SC_HASFIFO_DEFAULT	SC_HASFIFO_14
#endif

#ifdef IF_IODevice_Worked
#	define MuxMajor	[IODevice characterMajor]
#else
	static int	MuxMajor;
#endif
static void Muxintr ();

@implementation Mux

+ (BOOL) probe : description
{	Mux *			dev;
	IODirectDevice *	desc;

	/*
	 *	Find out what we have attached.
	 */
	dev = [self alloc];
	if (dev == nil) {
		IOLog ("%s: Failed to allocate an instance\n", MuxDeviceName);
		return NO;
	}

	if ([dev preInitHardware : description] == NO) {
		[dev free];
		return NO;
	}

	desc = [dev initFromDeviceDescription : description];
	if (desc == nil) {
		IOLog ("%s: Failed to init Device Description\n",MuxDeviceName);
		[dev free];
		return NO;
	}

	if ([dev postInitHardware : description  upper : self] == NO) {
		[dev free];
		return NO;
	}

	return YES;
}

- (BOOL) getUnsignedIntFromDescriptionString : description
				    argument : (unsigned int *)arg
					name : (char *)str;
{	const char *	String;
	IOConfigTable *	config = [description configTable];

	if (String = [config valueForStringKey : str]) {
		unsigned int	i;
		const char *	cp;

		for (cp = String, i = 0; ('0' <= *cp) && (*cp <= '9'); ++cp)
			i = (i * 10) + *cp - '0';
		[config freeString : String];
		/*
		 * We should check for an invalid number at this point, but
		 * then that would make the software somewhat critical with
		 * no apparent gains.
		 */
		*arg = i;
		return YES;
	}
	return NO;
}

- (unsigned) getInstance : description
{	unsigned	instance;

	if ([self getUnsignedIntFromDescriptionString : description
					     argument : &instance
						 name : "Instance"] == NO)
		instance = 0;
	return (instance);
}

/*
 *	This `C' code glue routine is used to convert an integer value into
 * the appropriate flags in sc_flags.
 */
static void
setFIFO (sp, val)
	struct tty_sc *	sp;
	int		val;
{
	sp->sc_flags &= ~(SC_HASFIFO_DEFAULT | SC_HASFIFO_1
	    | SC_HASFIFO_4 | SC_HASFIFO_8 | SC_HASFIFO_14);
	if ((val != 1) && (val != 4) && (val != 8) && (val != 14))
		sp->sc_flags |= SC_HASFIFO_DEFAULT;
	else if (val == 1)
		sp->sc_flags |= SC_HASFIFO_1;
	else if (val == 4)
		sp->sc_flags |= SC_HASFIFO_4;
	else if (val == 8)
		sp->sc_flags |= SC_HASFIFO_8;
	else if (val == 14)
		sp->sc_flags |= SC_HASFIFO_14;
}

- (BOOL) preInitHardware : description
{	unsigned int		unit, port, num, sum;
#	ifdef IF_IODevice_Worked
		IOReturn	ioReturn;
#	endif
	char			Name[sizeof(MuxDeviceName) + 1];
	unsigned		instance;
	static int		Muxopen(), Muxclose(), Muxread(), Muxwrite(),
				Muxselect(), Muxioctl();
	static void		Muxstop();
	extern			nulldev(), nodev();
	extern char *		strcpy();
	char const *		String;

#	ifdef DEBUG
		IOLog ("%s Initializing the instance\n", [self name]);
#	endif
	/*
	 *	An enhancement is to take the list of Interrupts and match
	 *	it up with the list of ports for a set of single ports ...
	 *	Currently if the port range is size of 1, it is the multiplexor
	 *	port. If the port range is size of 4, it will be an 8530. If
	 *	the port range is 8, then it is a 8250/16450/16550. The
	 *	multiplexor port must be the last port in the list.
	 */
	if ([description numPortRanges] < 1
	 || [description numInterrupts] != 1
	 || [description interrupt] > 15) {
		IOLog ("%s: Illegal parameters\n", MuxDeviceName);
		return NO;
	}
	if ([description interrupt] == 2) {
		int	aList[1];

		aList[0] = 9;
		[description setInterruptList : aList
					  num : 1];
	}
	(void)strcpy (Name, MuxDeviceName);
	Name[sizeof(Name) - 1] = '\0';
	Name[sizeof(Name) - 2] = (instance = [self getInstance : description])
	    + '0';
	[self setName : Name];
	[self setDeviceKind : MuxDeviceKind];
	[self setUnit : instance];
	if (instance >= NINSTANCE) {
		IOLog ("%s Too many instances\n", [self name]);
		return NO;
	}
	/*
	 * I use port here to suppress warning message about using
	 * uninitialized variable before setting it.
	 */
	num_devices[instance] = (port = [description numPortRanges]) - 1;

	if (instance == 0) {
#ifdef IF_IODevice_Worked
	    ioReturn = [self addToCdevswFromDescription : description
		open : Muxopen		 /* IOSwitchFunc openFunc	*/
		close: Muxclose		 /* IOSwitchFunc closeFunc	*/
		read : Muxread		 /* IOSwitchFunc readFunc	*/
		write: Muxwrite		 /* IOSwitchFunc writeFunc	*/
		ioctl: Muxioctl		 /* IOSwitchFunc ioctlFunc	*/
		stop : (int (*)())Muxstop /* IOSwitchFunc stopFunc	*/
		reset: nulldev		 /* IOSwitchFunc resetFunc	*/
		select: Muxselect	 /* IOSwitchFunc selectFunc	*/
		mmap : nodev		 /* IOSwitchFunc mmapFunc	*/
		getc : nodev		 /* IOSwitchFunc getcFunc	*/
		putc : nodev		 /* IOSwitchFunc putcFunc	*/];
	    if (ioReturn != YES) {
		IOLog ("%s addToCdevswFromDescription failed %s\n",
			[self name], [self stringFromReturn : ioReturn]);
		return NO;
	    }
#else
	    MuxMajor = IOAddToCdevsw (
		Muxopen,		 /* IOSwitchFunc openFunc	*/
		Muxclose,		 /* IOSwitchFunc closeFunc	*/
		Muxread,		 /* IOSwitchFunc readFunc	*/
		Muxwrite,		 /* IOSwitchFunc writeFunc	*/
		Muxioctl,		 /* IOSwitchFunc ioctlFunc	*/
		(int (*)())Muxstop,	 /* IOSwitchFunc stopFunc	*/
		nulldev,		 /* IOSwitchFunc resetFunc	*/
		Muxselect,		 /* IOSwitchFunc selectFunc	*/
		nodev,			 /* IOSwitchFunc mmapFunc	*/
		nodev,			 /* IOSwitchFunc getcFunc	*/
		nodev);			 /* IOSwitchFunc putcFunc	*/
#endif
	    if (MuxMajor == -1) {
		IOLog ("%s failed to place into Cdevsw\n", [self name]);
		return NO;
	    }
	}
#	ifdef DEBUG
		IOLog ("%s driver loaded on character device %d\n",
			[self name], MuxMajor);
#	endif
	/*
	 *	We intend to find out additional information about the devices
	 *	we have here! I allow for *automatic* sharing of interrupts (or
	 *	just to define ONE unmultiplexed port) to be allowed (typically
	 *	just a diode array) but just excluding the last port definition
	 *	for the Multiplexor port.
	 */
	if ([description portRangeList][port - 1].size == 8)
		++num_devices[instance];
	IOLog ("%s: probe for Serial Port Hardware irq %d:\n", [self name],
			    [description interrupt]);
	for (sum = unit = 0; unit < instance; ++unit)
		sum += num_devices[unit];
	first_sc[instance] = &sc[sum];
	for (num = unit = sum; unit < (sum + num_devices[instance]); ++unit) {
		int		i, oldcfcr, oldmsb;
		struct tty_sc *	sp;
		char		buf[32];

		port = ([description portRangeList][unit - sum]).start;

		IOLog ("\t0x%x: ", port);
		sp = &sc[num];

		(void)sprintf (buf, "Support CTS %u", num - sum);
		if ((!(String = [[description configTable]
		    valueForStringKey : buf])
		  && (!(String = [[description configTable]
		    valueForStringKey : "Support CTS"])))
		 || (strcmp (String, "YES") == 0)
		 || (strcmp (String, "yes") == 0))
			sp->sc_flags |= SC_HASCTS;

		/* Set the receive buffer size */
		(void)sprintf (buf, "Buffer Size %u", num - sum);
		if ((([self getUnsignedIntFromDescriptionString : description
		      argument : &sp->sc_size  name : buf] == NO)
		  && ([self getUnsignedIntFromDescriptionString : description
		      argument : &sp->sc_size  name : "Buffer Size"] == NO))
		 || (sp->sc_size > 65535)
		 || ((sp->sc_size < 16) && (sp->sc_size)))
			sp->sc_size = DOUBLE_BUFFER;

		/* Set the RTS Low Water Mark */
		(void)sprintf (buf, "RTS Low Water %u", num - sum);
		if (([self getUnsignedIntFromDescriptionString : description
		       argument : &sp->sc_rts_lowat  name : buf] == NO)
		  && ([self getUnsignedIntFromDescriptionString : description
		       argument : &sp->sc_rts_lowat  name : "RTS Low Water"]
		      == NO))
			sp->sc_rts_lowat = sp->sc_size / 3;
		if (sp->sc_rts_lowat > ((int)sp->sc_size - 2)) {
			if (sp->sc_size)
				sp->sc_rts_lowat = sp->sc_size / 3;
			else	sp->sc_rts_lowat = TTLOWAT (&(sp->sc_tty));
		}
		/* Set the RTS High Water Mark */
		(void)sprintf (buf, "RTS High Water %u", num - sum);
		if (([self getUnsignedIntFromDescriptionString : description
		       argument : &sp->sc_rts_hiwat  name : buf] == NO)
		  && ([self getUnsignedIntFromDescriptionString : description
		       argument : &sp->sc_rts_hiwat  name : "RTS High Water"]
		      == NO))
			sp->sc_rts_hiwat = (sp->sc_rts_lowat+sp->sc_size) >> 1;
		if (sp->sc_size) {
			if ((sp->sc_rts_hiwat > (sp->sc_size - 2))
			 || (sp->sc_rts_hiwat <= sp->sc_rts_lowat))
				sp->sc_rts_hiwat = (sp->sc_rts_lowat
				    + sp->sc_size) >> 1;
		} else	sp->sc_rts_hiwat = TTHIWAT (&(sp->sc_tty));

		/* Set the FIFO Water Mark */
		sp->sc_flags &= ~(SC_HASFIFO_DEFAULT | SC_HASFIFO_1
				| SC_HASFIFO_4 | SC_HASFIFO_8 | SC_HASFIFO_14);
		i = 0;
		(void)sprintf (buf, "FIFO %u", num - sum);
		if (([self getUnsignedIntFromDescriptionString : description
		       argument : &i  name : buf] == NO)
		 && ([self getUnsignedIntFromDescriptionString : description
		       argument : &i  name : "FIFO"]
		      == NO))
			sp->sc_flags |= SC_HASFIFO_DEFAULT;
		else	setFIFO (sp, i);

		/*
		 *	Clear any concept of an Interrupt Pending
		 *
		 *	Why?, Well, because we *hope* to add code to
		 *	automatically identify who interrupted us but can not
		 *	figure how to do this in the concept of an already
		 *	active kernel *with* interrupts already running from
		 *	multiple sources
		 */
		oldcfcr = inb (port + com_cfcr);/* Save old value of LINE_CTL */
		outb (port + com_cfcr, 0);	/* force access to id reg */
		outb (port + com_iir, 0);
		for (i = 0; i < 17; i++) {
			if ((inb (port + com_iir) & 0x38) == 0)
				break;			/* Yahoo! */
			(void)inb (port + com_data);	/* clear pending recv */
		}
		/* falling through, we may not have a leg to stand on */
		/* Save old value of DLBH */
		outb (port + com_cfcr, oldcfcr | CFCR_DLAB);
		oldmsb = inb (port + com_dlbh);
		/* Check if IER register writes through */
		outb (port + com_cfcr, oldcfcr &= ~CFCR_DLAB);
		outb (port + com_ier, 0);
		if (inb (port + com_ier) != 0) {
			outb (port + com_cfcr, oldcfcr | CFCR_DLAB);
			outb (port +com_dlbh, oldmsb);
			IOLog ("Port not installed (IER Failure)\n");
			continue;
		}
		/* Check if MSB of Divisor writes through */
		outb (port + com_cfcr, oldcfcr | CFCR_DLAB);
		outb (port + com_dlbh, 255);
		if (inb (port + com_dlbh) != 255) {
			outb (port + com_cfcr, oldcfcr | CFCR_DLAB);
			outb (port + com_dlbh, oldmsb);
			IOLog ("Port not installed (MDLAB Failure)\n");
			continue;
		}
		outb (port + com_cfcr, oldcfcr &~ CFCR_DLAB);
		if (inb (port + com_ier) != 0) {
			outb (port + com_cfcr, oldcfcr | CFCR_DLAB);
			outb (port + com_dlbh, oldmsb);
			IOLog ("Port not installed (Second Pass IER Failure)\n");
			continue;
		}
		/* Com port found, now check what chip it has */
		for (i = 0; i < 256; i++) {	 /* Is there Scratch register */
			outb (port + 7, i);
			if (inb (port + 7) != i)
				break;
		}
		/*
		 * It could have failed this test if the multiplexor board
		 * masks out the scratch register address ...
		 */
		if (i == 256) {			/* Yes == 450 or 460 */
			outb (port + 7, 0);
			i = 16450;
		} else
			i = 8250;
		/* look for a NS 16550AFN UART with FIFOs */
		outb (port + com_fifo, FIFO_ENABLE | FIFO_TRIGGER_14);
		DELAY (100);
		if ((inb (port + com_fifo) & IIR_FIFO_MASK) != 0) {
			if ((inb (port + com_fifo) & IIR_FIFO_MASK)
			    == IIR_FIFO_MASK) {
				sp->sc_flags |= SC_HASFIFO;
				switch (sp->sc_flags & (SC_HASFIFO_DEFAULT
				    | SC_HASFIFO_1 | SC_HASFIFO_4 | SC_HASFIFO_8
				    | SC_HASFIFO_14)) {
#				if (FIFO_TRIGGER) != (FIFO_TRIGGER_1)
				    case SC_HASFIFO_1:
					i = FIFO_ENABLE | FIFO_RCV_RST
					    | FIFO_XMT_RST | FIFO_TRIGGER_1;
					break;
#				endif
#				if (FIFO_TRIGGER) != (FIFO_TRIGGER_4)
				    case SC_HASFIFO_4:
					i = FIFO_ENABLE | FIFO_RCV_RST
					    | FIFO_XMT_RST | FIFO_TRIGGER_4;
					break;
#				endif
#				if (FIFO_TRIGGER) != (FIFO_TRIGGER_8)
				    case SC_HASFIFO_8:
					i = FIFO_ENABLE | FIFO_RCV_RST
					    | FIFO_XMT_RST | FIFO_TRIGGER_8;
					break;
#				endif
#				if (FIFO_TRIGGER) != (FIFO_TRIGGER_14)
				    case SC_HASFIFO_14:
					i = FIFO_ENABLE | FIFO_RCV_RST
					    | FIFO_XMT_RST | FIFO_TRIGGER_14;
					break;
#				endif
				default:
					i = FIFO_ENABLE | FIFO_RCV_RST
					    | FIFO_XMT_RST | FIFO_TRIGGER;
				}
				outb (port + com_fifo, i);
			} else {	/* Disable fifos */
				sp->sc_flags |= SC_BORKENFIFO;
				outb (port + com_fifo, 0);
			}
			i = 16550;
		}
		else	outb (port + com_fifo, 0);
		/* Lets see if it is the 82510 */
		if (i != 16550) {
			outb (port + com_fifo, FIFO_BANK_2);
			if (inb (port + com_fifo) & FIFO_BANK_2) {
				i = 82510;
				sp->sc_flags |= SC_HAS4FIFO;
				outb (port + com_fifo, FIFO_BANK_1);
				outb (port + com_rcm, FIFO_XMT_CLR);
				outb (port + com_tcm, FIFO_RCV_CLR);
			}
			outb (port + com_fifo, FIFO_BANK_0);
		}
		IOLog ("%s%d%s RTS %d<%d<%d%s\n",
		   (sc[num].sc_flags & SC_BORKENFIFO) ? "(Broken) " : "", i,
		   (sp->sc_size) ? " Buffered" : "",
		   sp->sc_rts_lowat, sp->sc_rts_hiwat,
		   (sp->sc_size) ? sp->sc_size : TTYHOG (&(sp->sc_tty)),
		   (sp->sc_flags & SC_HASCTS) ? " CTS" : "");
		DELAY (100);

#ifndef NO_MUX
		if ((i = [description portRangeList][
		    [description numPortRanges] - 1].size) <= 2) {
			sp->sc_mcr = MCR_INTROFF;
			if ((i != 2)	/* AST compatable Mux Port */
			 || (([description interrupt] & 1)
			  && ([description interrupt] != 9)))
				sp->sc_mux = [description portRangeList][
					[description numPortRanges] - 1].start;
			/*
			 * DigiBoard Mux port, if for an even interrupt, is
			 * the second port in the range.
			 */
			else	sp->sc_mux = [description portRangeList][
				     [description numPortRanges] - 1].start + 1;
			sp->sc_bit = unit - sum;
		} else
#endif
		/*
		 *	Single ports all with a common wired-or interrupt
		 *	line. There is no multiplex register in these simpler
		 *	cards (we detect no multiplex register specified by
		 *	noticing that the last port range is not 1 or 2 bytes).
		 */
		{	sp->sc_mcr = MCR_IENABLE;
#			ifndef NO_MUX
				sp->sc_mux = 0;
				sp->sc_bit = -1;
#			endif
			outb (port + com_mcr, sc[num].sc_mcr);
		}
		sp->sc_addr = port;
		++num;
	}
	if ((num - sum) == 1) {
		sc[sum].sc_mcr = MCR_IENABLE;
#		ifndef NO_MUX
			sc[sum].sc_mux = 0;
			sc[num].sc_bit = -1;
#		endif
		outb (port + com_mcr, sc[sum].sc_mcr);
	}
#	ifndef NO_MUX
		if (((num_devices[instance] = num - sum) > 1)
		    && (sc[sum].sc_mux))
			outb (sc[sum].sc_mux,
			    /* AST compatable Mux Port */
			    ([description portRangeList][
				[description numPortRanges] - 1].size != 2) ?
				    MUX_IENABLE_AST : MUX_IENABLE_DIGIBOARD) ;
#	endif
	[self registerDevice];
	return YES;
}

- (BOOL) postInitHardware : description
	upper : (id)probe
{
#	ifdef SerialPorts_TEST
		Mux *		dev;
#	endif

#	ifdef IF_IODevice_Worked
		IOReturn	ioReturn;

		if ((ioReturn = [self
		  enableInterrupt : [description interrupt]]) != IO_R_SUCCESS) {
			IOLog ("%s: Failed to set up Interrupt %d error %s\n",
			    MuxDeviceName, [description interrupt],
			    [self stringFromReturn : ioReturn]);
			return NO;
		}
#else
		unsigned	instance = [self getInstance : description];

		if ((instance + 1) > num_instance)
			num_instance = instance + 1;
		/*
		 *	This is the hack I use to workaround some Library
		 *	difficulties.
		 */
		(void)UNREGISTER_INTERRUPT([description interrupt]);
		if (REGISTER_INTERRUPT([description interrupt],
		 (IOEISAInterruptHandler *)Muxintr, instance, MUX_IPL) != YES) {
			IOLog ("%s: Failed to register IRQ %d\n",
			    MuxDeviceName, [description interrupt]);
			return NO;
		}
		if (ENABLE_INTERRUPT([description interrupt]) != YES) {
			IOLog ("%s: Failed to enable IRQ %d\n",
			    MuxDeviceName, [description interrupt]);
			return NO;
		}
#endif

	if (instance != 0)
		return YES;

#	ifdef SerialPorts_TEST
		/*
		 *	Spawn a SerialPorts version of the driver for
		 *	SerialMouse's pleasure.
		 */

		dev = [probe alloc];
		if (dev == nil) {
			IOLog (
			    "%s: Failed to allocate a SerialPorts instance\n",
			    MuxDeviceName);
			return YES;	/* Don't hamper our development */
		}

		[dev setName : "com0"];
		[dev setDeviceKind : MuxDeviceKind];
		[dev setUnit : instance];
		[dev registerDevice];
#	endif
	return YES;
}

#ifdef IF_IODevice_Worked
/*
 *	We wish to ensure that we are talking directly to the Interrupt
 *	occurrence since MACH interrupt handling is suspect because of Latency
 *	and loading issues (in other words, serial port processing is more
 *	important to me than disk servicing) ...
 */
- (BOOL) getHandler : (IOEISAInterruptHandler *)handler
	      level : (unsigned int *) ipl
	   argument : (unsigned in *) arg
       forInterrupt : (unsigned int) localInterrupt
{	unsigned	instance
		= [self getInstance : [self deviceDescription]];

#	ifdef DEBUG
		IOLog ("%s Enabling IRQ%d on Instance%d\n",
			[self name], localInterrupt, instance);
#	endif
	*handler = (IOEISAInterruptHandler)Muxintr;
	*ipl = IPLDEVICE;
	*arg = instance;
	return YES;
}
#endif

/*
 *	Name: getIntValues : values forParameter : parameter count : count
 *	Handle providing information back to user routines. This is an
 *	out-of-bound Kernel Server channel. We allow asking for the following:
 *		IOMajorDevice	Major device of this instance of the driver
 *		IOMinorDevice	First Minor device of this instance of driver.
 *		IONumberDevice	Number of ports handled on this instance.
 *		IOBufferSize	The default buffer size coded into the driver.
 *				This is not the real buffer size since this can
 *				be overridden with "Buffer Size" variables in
 *				the Instance?.table.
 *	The only services that use this following functionality is the
 *	SerialMouse driver and the SerialPorts PostLoader. We would like to
 *	allow getting the following (but have no documentation or clues to
 *	their actual purpose):
 *		SCC_CDEVSW_ENTRY  Major device of this instance of the driver
 *		SCC_WATCH_DCD
 *		SCC_WATCH_CTS
 *		SCC_WATCH_BREAK
 */
- (IOReturn) getIntValues : (unsigned *)values
	     forParameter : (IOParameterName) parameter
		    count : (unsigned *) count
{	unsigned	maxCount = *count;

	if (maxCount == 0) {
		maxCount = IO_MAX_PARAMETER_ARRAY_LENGTH;
	}
#	ifdef SerialPorts_TEST
		IOLog (
		    "%s getIntValues:values forParameter:\"%s\" count:%u\n",
		    [self name], parameter, maxCount);
#	endif
	if ((strcmp(parameter, "IOMajorDevice") == 0)
	 || (strcmp(parameter, "SCC_CDEVSW_ENTRY") == 0)) {
		values[0] = MuxMajor;
		*count = 1;
		return IO_R_SUCCESS;
	}

	if (strcmp (parameter, "IONumberDevice") == 0) {
		unsigned	instance
			= [self getInstance : [self deviceDescription]];

		values[0] = num_devices[instance];
		*count = 1;
		return IO_R_SUCCESS;
	}

	/*
	 *	Return the default Buffer Size.
	 */
	if (strcmp (parameter, "IOBufferSize") == 0) {
		values[0] = DOUBLE_BUFFER;
		*count = 1;
		return IO_R_SUCCESS;
	}

	/*
	 *	Return the Minor Device Number
	 */
	if (strcmp (parameter, "IOMinorDevice") == 0) {
		int	i, s;
		unsigned	instance
			= [self getInstance : [self deviceDescription]];

		for (i = s = 0; i < instance; ++i)
			s += num_devices[i];
		values[0] = s;
		*count = 1;
		return IO_R_SUCCESS;
	}

	return [super getIntValues : values
		      forParameter : parameter
			     count : &maxCount];
}

/*
 *	Name: setIntValues : values forParameter : parameter count : count
 *	Handle providing information to driver routines. This is an
 *	out-of-bound Kernel Server channel. The only services that use this
 *	functionality are the SerialMouse driver and the SerialPorts PostLoader.
 *	We would like to allow setting the following (but have no documentation
 *	or clues to their actual purpose):
 *		SCC_?_DATAFMT_PARAMS
 *		SCC_?_THRESHHOLD_PARAMS
 *		SCC_?_MODEM_DTR_PARAMS
 *		SCC_?_MODEM_RTS_PARAMS
 *		SCC_?_MODEM_BREAK_PARAMS
 */

- (IOReturn) setIntValues : (unsigned *)values
	     forParameter : (IOParameterName)parameter
		    count : (unsigned)count
{	unsigned	maxCount = count;

	if (maxCount == 0) {
		maxCount = IO_MAX_PARAMETER_ARRAY_LENGTH;
	}
#	ifdef DEBUG
		IOLog ("%s setIntValues:%d forParameter:\"%s\" count:%u\n",
			[self name], *values, parameter, maxCount);
#	endif
	if (strcmp (parameter, "SCC_A_DATAFMT_PARAMS") == 0) {
		int	i;

		i = 0;
		do {
			IOLog (" [%u]=%x", i, values[i]);
		} while (++i < 5);
		sc[0].sc_tty.t_ospeed = sc[0].sc_tty.t_ispeed = B1200;
		return IO_R_SUCCESS;
	}
	if (strcmp (parameter, "SCC_A_MODEM_DTR_PARAMS") == 0) {
		if (sc[0].sc_addr) {
			if (*values)
				sc[0].sc_mcr |= MCR_DTR;
			else	sc[0].sc_mcr &= ~MCR_DTR;
			outb (sc[0].sc_addr + com_mcr, sc[0].sc_mcr);
		} else
			IOLog ("SCC_A_MODEM_DTR_PARAMS on unopened port\n");
		return IO_R_SUCCESS;
	}
	if (strcmp (parameter, "SCC_A_MODEM_RTS_PARAMS") == 0) {
		if (sc[0].sc_addr) {
			if (*values)
				sc[0].sc_mcr |= MCR_RTS;
			else	sc[0].sc_mcr &= ~MCR_RTS;
			outb (sc[0].sc_addr + com_mcr, sc[0].sc_mcr);
		} else
			IOLog ("SCC_A_MODEM_DTR_PARAMS on unopened port\n");
		return IO_R_SUCCESS;
	}
	if (strncmp (parameter, "IOFIFO", 6) == 0) {
		int	which;
		char *	cp;
		unsigned	instance
			= [self getInstance : [self deviceDescription]];

		which = 0;
		if (*(cp = parameter+6) == '\0') {
			do {	if (sc[which+(instance*NTTY)].sc_flags
				    & SC_HASFIFO)
					setFIFO (&sc[which+(instance*NTTY)],
					    *values);
			} while (++which < NTTY);
		}
		else {	while (('0' <= *cp) && (*cp <= '9')) {
				which *= 10; which += *cp - '0';
			}
			if (sc[which+(instance*NTTY)].sc_flags & SC_HASFIFO)
				setFIFO (&sc[which+(instance*NTTY)], *values);
		}
		return IO_R_SUCCESS;
	}
	if (strncmp (parameter, "IOBufferSize", 12) == 0) {
		int	which;
		char *	cp;
		unsigned	instance
			= [self getInstance : [self deviceDescription]];

		which = 0;
		if (*(cp = parameter+12) == '\0') {
			do {
				sc[which+(instance*NTTY)].sc_size = *values;
			} while (++which < NTTY);
		}
		else {	while (('0' <= *cp) && (*cp <= '9')) {
				which *= 10; which += *cp - '0';
			}
			sc[which+(instance*NTTY)].sc_size = *values;
		}
		return IO_R_SUCCESS;
	}
	if (strncmp (parameter, "IOBufferHiwater", 15) == 0) {
		int	which;
		char *	cp;
		unsigned	instance
			= [self getInstance : [self deviceDescription]];

		which = 0;
		if (*(cp = parameter+15) == '\0') {
			do {
				sc[which+(instance*NTTY)].sc_rts_hiwat=*values;
			} while (++which < NTTY);
		}
		else {	while (('0' <= *cp) && (*cp <= '9')) {
				which *= 10; which += *cp - '0';
			}
			sc[which+(instance*NTTY)].sc_rts_hiwat = *values;
		}
		return IO_R_SUCCESS;
	}
	if (strncmp (parameter, "IOBufferLowater", 15) == 0) {
		int	which;
		char *	cp;
		unsigned	instance
			= [self getInstance : [self deviceDescription]];

		which = 0;
		if (*(cp = parameter+15) == '\0') {
			do {
				sc[which+(instance*NTTY)].sc_rts_lowat=*values;
			} while (++which < NTTY);
		}
		else {	while (('0' <= *cp) && (*cp <= '9')) {
				which *= 10; which += *cp - '0';
			}
			sc[which+(instance*NTTY)].sc_rts_lowat = *values;
		}
		return IO_R_SUCCESS;
	}
	return [super setIntValues : values
		      forParameter : parameter
			     count : maxCount];
}

#ifdef SerialPorts_TEST
/*
 *	Name: acquireUnit method
 *	Description: ?
 */
- (IOReturn) acquireUnit : (unsigned)unit
		     for : (IOParameterName)parameter
	      callbackId : (id)id
	       deviceTag : (unsigned)tag;
{
	IOLog ("acquireUnit:%u for:\"%s\" callbackId:%lx deviceTag:%u\n",
		unit, parameter, (long int)id, tag);
	return NO;
}

/*
 *	Name: bufferParameters method
 *	Description: ?
 */
- (IOReturn) bufferParameters
{
	IOLog ("bufferParameters\n");
	return NO;
}

/*
 *	Name: postReceiveBuffer method
 *	Description: ?
 */
- (IOReturn) postReceiveBuffer : (unsigned)unit
		 maxDataLength : (unsigned)length
		     bufferTag : (unsigned)tag
			vmTask : (id)id
		       forUnit : (unsigned)Unit;
{
	IOLog ("postReceiveBuffer:%u maxDataLength:%u bufferTag:%u vmTask:%u forUnit:%u\n",
		unit, length, tag, id, Unit);
	return NO;
}

/*
 *	Name: setStateEnable method
 *	Description: ?
 */
- (IOReturn) setStateEnable : (unsigned)unit
		    forUnit : (unsigned)Unit;
{
	IOLog ("setStateEnable:%u forUnit:%u\n",
		unit, Unit);
}
#endif

/*
 *	No More Objects after this point, I am now a regular human :-)
 */

#define IS_LOCAL(sp)	(sp->sc_nty->t_pflags & TP_CLOCAL)
#define MUX_LOCAL(x)	((minor(x)&(128|64))!=64)
/* Special new device that ignores DCD on open, but will detect a hangup */
#define MUX_HANGUP(x)	((minor(x)&(128|64))==128)
#define MUX_LOCK(x)	(minor(x)&64)
/* This assumes that `sp' is available */
#define MUX_FLOW(x)	((minor(x)&32) && (sp->sc_flags & SC_HASCTS))
#define	MUX_UNIT(x)	(minor(x)&31)
extern void	wakeup();
/* to ensure we don't run too many concurrent copies in the timeout queue */

static unsigned		doneMuxRxExecutive;
#define notMuxRxExecutive()	(++doneMuxRxExecutive == 1)
static unsigned		doneMuxTxExecutive;
#define notMuxTxExecutive()	(++doneMuxTxExecutive == 1)

/*
 *	This routine is called via the timeout queue to process receive
 *	interrupts. Flush the Double Buffer into the CLIST buffer
 */
static void
MuxRxExecutive (dev)
	int		dev;
{	struct tty_sc *	sp, *st;
	struct tty *	tp;
	int		num, n, keep_going, keep_trying, port, mcr;
	short *		rp;

	do {
	    doneMuxRxExecutive = 1;
	    if (dev >= 0) {
		st = &sc[MUX_UNIT ((dev_t)dev)];
		n = 1;
	    } else {
		st = sc;
		for (n = num = 0; num < num_instance; ++num)
		    n += num_devices[num];
	    }
	    dev = -1;
	    do {
		keep_going = keep_trying = 0;
		for (sp = st, num = n; num > 0; --num, ++sp) {
		    if (sp->sc_size == 0)
			continue;
		    tp = &(sp->sc_tty);
		    port = sp->sc_addr;
		    mcr = -1;
		    if ((tp->t_state & TS_ISOPEN) && (sp->sc_rxbuf)) {
			int	parenb = (
			 ((sp->sc_nty->t_pflags
			 & (TP_PARENB|TP_IGNPAR))== (TP_PARENB)) &&
			  (((tp->t_flags & ANYP) == EVENP)
	 		  || (tp->t_flags & ANYP) == ODDP));
			int	overflow = 0;
			int	code, c;

			rp = sp->sc_rxout;
			while ((sp->sc_rxin != rp) && sp->sc_rxbuf) {
			    if ((tp->t_rawq.c_cc + tp->t_canq.c_cc)
			      >= TTYHOG (tp)) {
				++keep_trying ;
				break;
			    }
			    c = *((unsigned char *)rp);
			    code = *(((char *)rp) + 1);
			    if (code & (LSR_BI | LSR_FE))
				c |= TTY_FE;
			    else if ((code & LSR_PE) && parenb)
				c |= TTY_PE;
			    if (code & LSR_OE)
				overflow = 1;
			    /*
			     *	If we receive a
			     * character we can reset the
			     * modem status change counter.
			     */
			    sc->sc_mcount = 0;
			    (*linesw[(int)tp->t_line].l_rint) (c, tp);
			    if (sp->sc_rxbuf == (short *)0L)
				break;
			    if (++rp >= TOP_RXBUF(sp))
				rp = sp->sc_rxbuf;
			    sp->sc_rxout = rp;
			    --(sp->sc_rxnum);
			    ++keep_going;
			    /*
			     *	When our size goes below the buffer watermark,
			     * assert RTS. We optimize somewhat since an output
			     * call takes a considerable period of time.
			     */
			    if (port && (mcr != (c = MCR (sp))))
				    outb (port + com_mcr, mcr = c);
			}
			/*
			 *	Log a warning message, the user
			 * should make this a hint that they
			 * are exceding the throughput
			 * capability of the software.
			 */
			if (overflow)
			    IOLog("%s[%x]: silo overflow\n",
				MuxDeviceName, tp->t_dev);
		    } else {sp->sc_rxout = sp->sc_rxin;
			    sp->sc_rxnum = 0;
		    }
		}
	    } while (keep_going);
	    n = splmux ();
	    num = --doneMuxRxExecutive;
	    splx (n);
	} while (num != 0);
	if (keep_trying) {
	    n = splmux ();
	    num = notMuxRxExecutive();
	    splx (n);
	    if (num)
		ns_timeout ((func)MuxRxExecutive, (void *)-1L,
		    (ns_time_t)100000000ULL, CALLOUT_PRI_SOFTINT0);
	}
}

extern void	selwakeup();

/*
 *	This routine is called via the timeout queue to process transmit
 *	interrupts. Flush the CLIST buffer into the Transmitter buffer.
 */
static void
MuxTxExecutive (dev)
	int		dev;
{	struct tty_sc *	sp, *st;
	struct tty *	tp;
	int		num, n;
	char *		cp;
	static void	Muxstart();

	do {
	    doneMuxTxExecutive = 1;
	    if (dev >= 0) {
		st = &sc[MUX_UNIT ((dev_t)dev)];
		n = 1;
	    } else {
		st = sc;
		for (n = num = 0; num < num_instance; ++num)
		    n += num_devices[num];
	    }
	    dev = -1;
	    for (sp = st, num = n; num > 0; --num, ++sp) {
		if (sp->sc_size == 0)
		    continue;
		tp = &(sp->sc_tty);

		if ((tp->t_state & TS_ISOPEN) && (sp->sc_txbuf)) {
		    /*
		     *	If the buffer is not full, lets start ripping out as
		     * much of the output CLIST as we can (while we still have a
		     * task). The interrupt should not change sc_txin.
		     */
		    if (sp->sc_txbuf != NULL) {
			if (tp->t_outq.c_cc > 0) {
			    do {
				if ((cp = sp->sc_txin + 1) >= TOP_TXBUF(sp))
				    cp = sp->sc_txbuf;
				if (cp == sp->sc_txout)
				    break;
				*(sp->sc_txin) = getc(&tp->t_outq);
				sp->sc_txin = cp;
			    } while (tp->t_outq.c_cc > 0);
			}
		    } else
			while (tp->t_outq.c_cc > 0)
			    getc (&tp->t_outq);
		    if (tp->t_outq.c_cc <= TTLOWAT(tp)) {
			if (tp->t_state & TS_ASLEEP) {
			    tp->t_state &= ~TS_ASLEEP;
			    wakeup ((caddr_t) & tp->t_outq);
			}
			if (tp->t_wsel) {
			    selwakeup (tp->t_wsel, tp->t_state & TS_WCOLL);
			    tp->t_wsel = 0;
			    tp->t_state &= ~TS_WCOLL;
			}
		    }
		    /*
		     *	Lets call the start routine to possibly circumvent
		     * a transmit interrupt.
		     */
		    Muxstart (tp);
		    --doneMuxTxExecutive;	/* Incremented by above */
		} else
		    sp->sc_txout = sp->sc_txin;
	    }
	    n = splmux ();
	    num = --doneMuxTxExecutive;
	    splx (n);
	} while (num != 0);
}

static void
Muxstart (tp)
	struct tty *	tp;
{	struct tty_sc *	sp;
	int		unit, port, flag, s, i;
	char *	cp;

	/* schedule to refill the transmit buffer if need be */
	unit = MUX_UNIT (tp->t_dev);
	sp = &sc[unit];
	if (sp->sc_size) {
		s = splmux ();
		i = notMuxTxExecutive();
		splx (s);
		if (i)
		    ns_timeout ((func)MuxTxExecutive, (void *)-1L,
			(ns_time_t)1ULL, CALLOUT_PRI_SOFTINT0);
	} else {
		if (tp->t_outq.c_cc <= TTLOWAT(tp)) {
			if (tp->t_state & TS_ASLEEP) {
			    tp->t_state &= ~TS_ASLEEP;
			    wakeup ((caddr_t) & tp->t_outq);
			}
			if (tp->t_wsel) {
			    selwakeup (tp->t_wsel, tp->t_state & TS_WCOLL);
			    tp->t_wsel = 0;
			    tp->t_state &= ~TS_WCOLL;
			}
		}
	}
	if (((sp->sc_size)
	    ? (sp->sc_txin == sp->sc_txout) : (tp->t_outq.c_cc == 0))
	 || (tp->t_state & (TS_TIMEOUT | TS_TTSTOP)))
		return;
	s = splmux ();
	/*
	 *	Idealy I would like to fill in all the serial ports on this
	 *	same Multiplex interrupt so as to ensure that they are all
	 *	synchronized. This reduces the interrupt load on the machine
	 *	considerably since all the transmitters will empty out at
	 *	roughly the same time.
	 */
	if ((port = sp->sc_addr) && (inb (port + com_lsr) & LSR_TXRDY)) {
		if (MUX_FLOW (tp->t_dev) && !(inb (port + com_msr) & MSR_CTS)) {
			tp->t_state |= TS_TTSTOP;		/* for pstat */
			splx (s);
			return;
		}
		tp->t_state |= TS_BUSY | TS_OUTPUTBUSY;
		i = 1;
		flag = sp->sc_flags;
		if (flag & SC_HASFIFO)
			i = 16;		/* Wheeeeeee */
		else if (flag & SC_HAS4FIFO)
			i = 4;
		if (sp->sc_size)
		for (; i && ((cp = sp->sc_txout) != sp->sc_txin); i--) {
			outb (port + com_data, *cp);
			if (++cp >= TOP_TXBUF(sp))
				cp = sp->sc_txbuf;
			sp->sc_txout = cp;
		}
		/*
		 *	If no double buffering, get the data directly
		 */
		else for (; i && tp->t_outq.c_cc; i--)
			outb (port + com_data, getc (&tp->t_outq));
		/*
		 *	If we transmit a character we can reset the modem
		 * status change counter.
		 */
		sc->sc_mcount = 0;
	}
	splx (s);
}

static int
Muxspeed (spd)
	long spd;
{	long x, err, speed;
#define divrnd(n, q)	((((n) << 1) / (q) + 1) >> 1) /* divide and round off */

	switch (spd) {
	default:	speed = spd;	break;
	case B0:	return (0);
	case B50:	speed = 50;	break;
	case B75:	speed = 75;	break;
	case B110:	speed = 110;	break;
	case B134:	speed = 134;	break;
	case B150:	speed = 150;	break;
	case B200:	speed = 200;	break;
	case B300:	speed = 300;	break;
	case B600:	speed = 600;	break;
	case B1200:	speed = 1200;	break;
	case B1800:	speed = 1800;	break;
	case B2400:	speed = 2400;	break;
	case B4800:	speed = 4800;	break;
	case B9600:	speed = 9600;	break;

#	ifdef B14400
	  case B14400:	speed = 14400;	break;
#	endif

#	ifdef B19200
	  case B19200:	speed = 19200;	break;
#	else
	  case EXTA:	speed = 19200;	break;
#	endif

#	ifdef B28800
	  case B28800:	speed = 28800;	break;
#	endif

#	ifdef B38400
	  case B38400:	speed = 38400;	break;
#	else
	  case EXTB:	speed = 38400;	break;
#	endif

#	ifdef B43200
	  case B43200:	speed = 43200;	break;
#	endif

#	ifdef B57600
	  case B57600:	speed = 57600;	break;
#	endif

/* Ya sure, and I also know the secret to the Universe! */
#	ifdef B115200
	  case B115200:	speed = 115200;	break;
#	endif
	}

	if (speed == 0)
		return (0);
	if (speed < 0)
		return (-1);

#	ifdef DEBUG
		IOLog ("%s: calculated speed: %d\n", MuxDeviceName, speed);
#	endif
	x = divrnd (COMTICK, speed);
	if (x <= 0)
		return (-1);

#	ifdef SPEED_TOLERANCE
		err = divrnd(1000 * COMTICK, x * speed) - 1000;
		if (err < 0)
			err = -err;
		if (err > SPEED_TOLERANCE)
			return (-1);
#	endif
	return (x);

#undef divrnd
}

#define t_wopen	t_state

static int
Muxparam (unit)
	int		unit;
{	struct tty_sc *	sp = &sc[unit];
	struct tty *	tp = &sp->sc_tty;
	int		port = sp->sc_addr;
	int		cfcr, flag;
	int		s, ospeed = Muxspeed(tp->t_ospeed);
	long		lflags;

	/* check requested parameters */
	if ((port == 0) || (ospeed < 0)
	 || (tp->t_ispeed && tp->t_ispeed != tp->t_ospeed))
		return (EINVAL);
	flag = tp->t_flags;
	lflags = sp->sc_nty->t_pflags;
	/*
	 *	I think I want to do this, but needs more work
	 *
	 *	if (MUX_LOCK(tp->t_dev) && (lflags & TP_CLOCAL)
	 *	 && (tp->t_state & TS_WOPEN))
	 *		wakeup((caddr_t) &tp->t_wopen);
	 */
	/* If our speed is 0, then we are in the HUPCLS state */
	if (ospeed == 0) {
		tp->t_state |= TS_HUPCLS;
		/* hang up line */
		outb (port + com_mcr, sp->sc_mcr &= ~MCR_DTR);
		return (0);
	}
	s = splmux ();
	outb (port + com_cfcr, inb (port + com_cfcr) & ~CFCR_DLAB);
	/*
	 *	It appears that the Mach 2.5 Kernel, with the NeXTSTEP hacks
	 * to handle the *new* tty structure has a great deal of trouble
	 * reconciling the CREAD flag. I will chose to ignore it's value and
	 * leave the receiver interrupt active, despite the savings it may
	 * incure for the selective software.
	 *	outb (port + com_ier, (lflags & TP_CREAD)
	 *	 ? IER_ERXRDY | IER_ETXRDY | IER_ERLS | IER_EMSC
	 *	 : IER_ETXRDY | IER_EMSC);
	 */
	/*
	 *	If we *need* any information from the Modem Control Lines,
	 *	then enable that interrupt, otherwise, do not *overservice*
	 *	us if we have no purpose for a particularily noisy device
	 *	connected to DCD or RTS.
	 */
	outb (port + com_ier, (
		((!MUX_LOCAL(tp->t_dev)) || MUX_FLOW(tp->t_dev)
		 || (tp->t_state & TS_WOPEN))
		? (IER_ERXRDY | IER_ETXRDY | IER_ERLS | IER_EMSC)
		: (IER_ERXRDY | IER_ETXRDY | IER_ERLS)));
	outb (port + com_cfcr, inb (port + com_cfcr) | CFCR_DLAB);
	outb (port + com_dlbl, ospeed & 0xFF);
	outb (port + com_dlbh, ospeed >> 8);
	cfcr = CFCR_7BITS;
	if (flag & (RAW | PASS8 | PASS8OUT | LITOUT))
		cfcr = CFCR_8BITS;
	else if ((flag & CBREAK) == 0) switch (lflags & TP_CSIZE) {
	case TP_CS5:	cfcr = CFCR_5BITS;	break;
	case TP_CS6:	cfcr = CFCR_6BITS;	break;
     /* case TP_CS7:	cfcr = CFCR_7BITS;	break; */
	case TP_CS8:	cfcr = CFCR_8BITS;	break;
	}
	/* Is this right? We will disable parity if RAW */
	if ((lflags & TP_PARENB)
	 && !(flag & (RAW | PASS8 | PASS8OUT | LITOUT))) {
		cfcr |= CFCR_PENAB;
		switch (flag & ANYP) {
		case EVENP:	cfcr |= CFCR_PEVEN;	break;
		case ODDP:	cfcr |= CFCR_PODD;	break;
		case EVENP|ODDP:cfcr |= CFCR_PONE;	break;
		case 0:		cfcr |= CFCR_PZERO;	break;
		}
	}
	/*
	 * 110 baud uses two stop bits - all other speeds use one, unless ...
	 */
	if (((lflags & TP_CSTOPB110) && (tp->t_ispeed == B110))
	 || (lflags & TP_CSTOPB))
		cfcr |= CFCR_STOPB;
#	if (defined(DEBUG))
		IOLog ("Muxparam[%x]: flag=%x lflags=%lx cfcr=%x\n",
		    unit, flag, lflags, cfcr);
#	endif
	outb (port + com_cfcr, cfcr);
	/* if the previous speed was 0, need to re-enable DTR? */
	outb (port + com_mcr, MCR (sp));

	if (sc[unit].sc_flags & SC_HASFIFO) {
		switch (sp->sc_flags & (SC_HASFIFO_DEFAULT | SC_HASFIFO_1
		    | SC_HASFIFO_4 | SC_HASFIFO_8 | SC_HASFIFO_14)) {
#		if (FIFO_TRIGGER) != (FIFO_TRIGGER_1)
		    case SC_HASFIFO_1:
			outb (port + com_fifo, FIFO_ENABLE | FIFO_TRIGGER_1);
			break;
#		endif
#		if (FIFO_TRIGGER) != (FIFO_TRIGGER_4)
		    case SC_HASFIFO_4:
			outb (port + com_fifo, FIFO_ENABLE | FIFO_TRIGGER_4);
			break;
#		endif
#		if (FIFO_TRIGGER) != (FIFO_TRIGGER_8)
		    case SC_HASFIFO_8:
			outb (port + com_fifo, FIFO_ENABLE | FIFO_TRIGGER_8);
			break;
#		endif
#		if (FIFO_TRIGGER) != (FIFO_TRIGGER_14)
		    case SC_HASFIFO_14:
			outb (port + com_fifo, FIFO_ENABLE | FIFO_TRIGGER_14);
			break;
#		endif
		default:
			outb (port + com_fifo, FIFO_ENABLE | FIFO_TRIGGER);
		}
	} else if (sc[unit].sc_flags & SC_HAS4FIFO) {
		outb (port + com_fifo, FIFO_BANK_2);
		outb (port +com_idm, FIFO_ENBL82510);
		outb (port + com_fifo, FIFO_BANK_1);
		outb (port + com_tcm, FIFO_XMT_CLR);
		outb (port + com_rcm, FIFO_RCV_CLR);
		outb (port + com_fifo, FIFO_BANK_2);
		outb (port + com_idm, FIFO_CLEAR_CMD);
		outb (port + com_fifo, FIFO_BANK_0);
	}
	splx (s);
	return (0);
}

/* ARGSUSED */
static int
Muxopen (dev_t dev, int flag, int mode)
{	int		i, s, error = 0;
	int		unit = MUX_UNIT (dev);
	struct tty_sc *	sp = &sc[unit];
	int		port = sp->sc_addr;
	struct tty *	tp = &sp->sc_tty;

	for (s = i = 0; i < num_instance; ++i)
		s += num_devices[i];
	if ((unit >= s) || (port == 0))
		return (ENXIO);
	tp->t_oproc = (int (*)())Muxstart;
	if ((tp->t_state & (TS_ISOPEN | TS_WOPEN)) == 0) {
		extern int	ttychars ();

		ttychars (tp);
		tp->t_addr = (caddr_t)port;	/* no purpose, registration */
		/*
		 *	This just makes sure we have all the pointers
		 *	making sense, even if the rxbuf is dynamically
		 *	allocated ...
		 */
		sp->sc_rxin = sp->sc_rxout = sp->sc_rxbuf;
		sp->sc_rxnum = 0;
		sp->sc_txin = sp->sc_txout = sp->sc_txbuf;
		/* We need some defaulting ... */
		if (tp->t_ispeed == 0) {
			struct proc *	p = getproc() ;
#			if NeXT
				/* struct utask *	up ; */
#			endif

			sp->sc_nty = ttynty (tp);
			NTYDEFAULTS (sp->sc_nty);
			sp->sc_nty->t = tp ;
#			if NeXT
				/* if ((up = u_task()) && !(up->uu_ttyd)) {
					up->uu_ttyd = dev ;
					up->uu_ttyp = tp ;
					p->p_flag |= SCTTY ;
				} */
#			else
				if (!(p->p_ttyd)) {
					p->p_ttyd = dev ;
					p->p_ttyp = tp ;
					p->p_flag |= SCTTY ;
				}
				if (!(p->p_logdev))
					p->p_logdev = dev ;
#			endif
			if (!(sp->sc_nty->t_posix_pgrp))
				sp->sc_nty->t_posix_pgrp = pgfind (p->p_pid);
			tp->t_ispeed = tp->t_ospeed = B9600;
			tp->t_flags = ANYP | ECHO | CRMOD;
		}
		Muxparam (unit);
		tp->t_state &= ~(TS_BUSY | TS_OUTPUTBUSY);
	}
	/*
	 *	Great, if I get what I want, no root access to the port even,
	 *	if both ports trying are locking devices and opposite sides.
	 *		} else if (tp->t_state&TS_XCLUDE && (!suser()
	 *		 || (MUX_LOCK(tp->t_dev) && MUX_LOCK(dev)
	 *		  && ((!MUX_LOCAL(tp->t_dev) && MUX_LOCAL(dev))
	 *		   || (MUX_LOCAL(tp->t_dev) && !MUX_LOCAL(dev))))))
	 *	Instead, lets just match up the Mach task that opened the port,
	 *	and call it a day.
	 */
	else if (((tp->t_state & TS_XCLUDE) && (sp->sc_task != current_task ())
	  && !suser ())
	 || ((MUX_LOCK (tp->t_dev) || MUX_LOCK (dev)) && (tp->t_dev != dev)
	  && (tp->t_dev)
	  && (MUX_LOCAL (tp->t_dev) || ((tp->t_state & TS_WOPEN) == 0)))) {
#		if (defined (DEBUG))
			IOLog("%s[%x]: EBUSY %s%s%s%s\n", MuxDeviceName,
			    sp->sc_tty.t_dev,
			    (tp->t_state & TS_XCLUDE) ? "TS_XCLUDE, " : "",
			    (sp->sc_task != current_task()) ? "TASK, " : "",
			    (suser ()) ? "SUSER, " : "",
			    (tp->t_state & TS_WOPEN) ? "TS_WOPEN, " : "");
#		endif
		return (EBUSY);
	}
	if (MUX_LOCAL (dev))
		sp->sc_nty->t_pflags |= TP_CLOCAL;
	else	sp->sc_nty->t_pflags &= ~TP_CLOCAL;
	s = splmux ();
	sp->sc_mcr |= MCR_DTR;
	outb (port + com_mcr, MCR (sp));
	DELAY (1000);
	if (((inb (port + com_msr)) & MSR_DCD)
	 || MUX_LOCAL (dev) || IS_LOCAL (sp))
		tp->t_state |= TS_CARR_ON;
	else	tp->t_state &= ~TS_CARR_ON;
	if (!MUX_LOCAL (dev)) {
		/*
		 *	Beware of a Re-open ...
		 */
		while (((tp->t_state & TS_XCLUDE)
		/*
		 * Christian Starkjohann feels this following is necessary for
		 * some software to operate. This does *not* follow the
		 * correct standard for XCLUDE mode and could represents a
		 * security risk as well ... :-(
		 */
		  && (sp->sc_task != current_task ()))
		 || ((tp->t_state & TS_CARR_ON) == 0)
		 || ((MUX_LOCK (tp->t_dev) || (MUX_LOCK (dev)))
		  && (tp->t_dev != dev) && (tp->t_dev))) {
			extern int	tsleep();

			sp->sc_mcr |= MCR_DTR;
			outb (port + com_mcr, MCR (sp));
			tp->t_state |= TS_WOPEN;
			/* Muxparam may shut down interrupts in a race */
			outb (port + com_ier,
			    IER_ERXRDY | IER_ETXRDY | IER_ERLS | IER_EMSC);
			splx (s);
			if (error = tsleep ((caddr_t)&tp->t_wopen,
			    TTIPRI | PCATCH,
			    /* ttyopen ?? */
			    linesw[(int)tp->t_line].l_open, 0)) {
				s = splmux ();
				break;
			}
			/*
			 * 100ms DTR is to be dropped on rewake (close).
			 * Thanks for Dennis Breckenridge for pointing this one
			 * out, most modems will observe the DTR signal if it
			 * drops for more than 50ms, we will wait 100ms.
			 */
			IOSleep (100);
			s = splmux ();
			if (MUX_LOCK (tp->t_dev) == 0) {
				if (tp->t_ispeed == 0) {
					sp->sc_nty = ttynty (tp);
					NTYDEFAULTS (sp->sc_nty);
					tp->t_ispeed = tp->t_ospeed = B9600;
					tp->t_flags = ANYP | ECHO | CRMOD;
				}
				/* Was cleared by Muxclose, for Muxparam */
				tp->t_state |= TS_WOPEN;
				/* Re-enable Interrupts */
				Muxparam (unit);
			}
		}
		sp->sc_mcr |= MCR_DTR;
		outb (port + com_mcr, MCR (sp));
		tp->t_state &= ~TS_WOPEN;
	}
	/* This has the side affect of locking the port if it is meant to */
	tp->t_dev = dev;
	if (sp->sc_size && (error == 0) && (sp->sc_kalloc == (char *)0L)) {
		extern void *	kalloc ();

		sp->sc_ksize = (
		    (sp->sc_size * sizeof(short))	/* Receive Buffer */
		  + (sp->sc_size * sizeof(char))	/* Transmit Buffer */
			+ sizeof(short));		/* Alignment splunge */
		if (sp->sc_kalloc = (char *)kalloc (sp->sc_ksize)) {
			short *	rp;

			/*
			 * Dangerous to set rxbuf first because interrupts and
			 * Executives could get in and start fiddling (I
			 * know, we are at MUX_IPL, but hitting memory is panic
			 * territory ...)
			 */
			/* Align on short boundary */
			rp = (short *)((((long)sp->sc_kalloc)
			    | ((1 << (sizeof(short) - 1)) - 1)) + 1);
			/* sp->sc_rts_lowat = sp->sc_size / 3;		*/
			/* sp->sc_rts_hiwat = (sp->sc_size << 1) / 3;	*/
			sp->sc_rxtop = rp + sp->sc_size;
			sp->sc_rxin = sp->sc_rxout = rp;
			sp->sc_rxnum = 0;
			sp->sc_rxbuf = rp;
			sp->sc_txtop = ((char *)(sp->sc_rxtop)) + sp->sc_size;
			sp->sc_txin = sp->sc_txout = (char *)(sp->sc_rxtop);
			sp->sc_txbuf = (char *)(sp->sc_rxtop);
		} else {
			error = ENOMEM;
			sp->sc_kalloc = sp->sc_txbuf = NULL;
			sp->sc_rxbuf = (short *)NULL;
		}
	}
	if (error == 0)
		sp->sc_task = current_task ();
	splx (s);
	if (error == 0)
		error = (*linesw[(int)tp->t_line].l_open) (dev, tp);
	return (error);
}

/*ARGSUSED*/
static int
Muxclose (dev, flag, mode)
	dev_t		dev;
	int		flag, mode;
{	int		unit = MUX_UNIT (dev);
	struct tty_sc *	sp = &sc[unit];
	struct tty *	tp = &sp->sc_tty;
	int		port;
	int		wopen = tp->t_state & TS_WOPEN;
	extern void	ttyclose ();

	/* We are closing it, doesn't matter now about holding back ... */
	tp->t_state &= ~TS_TTSTOP;
	(*linesw[(int)tp->t_line].l_close) (tp, flag);
	if (port = sp->sc_addr) {
		outb (port + com_cfcr, inb (port + com_cfcr) &~
		    (CFCR_SBREAK | CFCR_DLAB));
		/*
		 * Disable Interrupts, the wakeup below will re-enable them
		 * if necessary
		 */
		if ((tp->t_state & TS_HUPCLS) || wopen
		 || ((tp->t_state & TS_ISOPEN) == 0)) {
			/*
			 *	Since the line is set up for HUPCL, and it is
			 * the last close for this line, then we can set the
			 * speed to 0 to hang up the line.
			 */
			tp->t_ispeed = 0;
			if (wopen == 0)
				outb (port + com_ier, 0);
			outb (port + com_mcr, sp->sc_mcr &= ~MCR_DTR);
		}
	}
	/* Local only if we set explicitly, or if we open the correct device */
	sp->sc_nty->t_pflags &= ~TP_CLOCAL;
	/* This flag should be cleared by ttyclose, but I don't trust mach */
	tp->t_state &= ~TS_XCLUDE;
	sp->sc_task = (task_t)0L;
	ttyclose (tp);
	/*
	 *	Deallocate any buffers we had
	 */
	if ((wopen == 0) && (sp->sc_kalloc)) {
		extern void	kfree ();

		/*
		 *	First, we switch off the buffer (in case either an
		 *	interrupt or Executive task pops in), then we release
		 *	the storage.
		 */
		sp->sc_rxbuf = (short *)0L;
		sp->sc_txin = sp->sc_txout = sp->sc_txbuf = (char *)0L;
		kfree (sp->sc_kalloc, sp->sc_ksize);
		sp->sc_kalloc = (char *)0L;
	}
	tp->t_dev = 0;	/* Unlock port */
	/* Wake up sleeping opens in case they are waiting on lock */
	wakeup ((caddr_t) &tp->t_wopen);
	return (0);
}

/*
 *	This entry point is used by the kernel to extract data from the Clist
 *	and place it into the user space. The line discipline looks after any
 *	of the vagaraties of the post processing. It may be desired to handle
 *	any forms of double buffering here, and to *poll* the actual serial
 *	receive port in cases of impending context switches.
 */
static struct tty *
Muxfillread (dev)
	dev_t		dev;
{	int		s, i;
	struct tty_sc *	sp = &sc[MUX_UNIT (dev)];
	struct tty *	tp = &sp->sc_tty;

	/*
	 *	I have no idea if this improves throughput, but at
	 * least with the double buffered approach I can implement RTS
	 * handshaking and reliable 38400 operation. It is up to the
	 * user of this driver to determine if this is good for their
	 * particular operation.
	 */

	/*
	 *	It is far better, at this point, if we are
	 * running a port with a FIFO to check the FIFO
	 * *before* chancing going to sleep with the
	 * accompanying context switch.
	 */
	if ((sp->sc_flags & SC_HASFIFO)
	 && ((sp->sc_flags & (SC_HASFIFO_1 | SC_HASFIFO_4 | SC_HASFIFO_8
	     | SC_HASFIFO_14)) != SC_HASFIFO_1)
	 && (tp->t_rawq.c_cc <= 0)	/* No characters? */
	 && (inb (sp->sc_addr+com_lsr) & LSR_RCV_MASK)) {
		int	instance;

		for (instance = 0; (instance < num_instance)
		 && (sp < first_sc[instance]); ++instance);
		s = splmux ();
#		ifdef IF_IODevice_Worked
			Muxintr((void *)NULL, (void *)NULL, instance);
#		else
			Muxintr(instance);
#		endif
		splx (s);
	}
	/* already in the queue ? */
	if (sp->sc_size) {
		s = splmux ();
		i = notMuxRxExecutive();
		splx (s);
		if (i)
			MuxRxExecutive ((int)dev);
	}
	return (tp);
}

static int
Muxread (dev, uio, flag)
	dev_t		dev;
	struct uio *	uio;
	int		flag;
{	struct tty *	tp;

	/* schedule to fill the read buffer if need be */
	tp = Muxfillread (dev);
	return (*linesw[(int)tp->t_line].l_read) (tp, uio, flag);
}

static int
Muxwrite (dev, uio, flag)
	dev_t		dev;
	struct uio *	uio;
	int		flag;
{	struct tty_sc *	sp = &sc[MUX_UNIT (dev)];
	struct tty *	tp = &sp->sc_tty;
	int		ret, s, i;

	ret = (*linesw[(int)tp->t_line].l_write) (tp, uio, flag);
	/* schedule to refill the transmit buffer if need be */
	if (sp->sc_size) {
		s = splmux ();
		i = notMuxTxExecutive();
		splx (s);
		if (i)
		    ns_timeout ((func)MuxTxExecutive, (void *)-1L,
			(ns_time_t)1ULL, CALLOUT_PRI_SOFTINT0);
	}
	return (ret);
}

static int
Muxselect (dev, rw)
	dev_t		dev;
	int		rw;
{	struct tty *	tp;

	tp = Muxfillread (dev);
	return ((*linesw[(int)tp->t_line].l_select) (tp, rw));
	/* tpselect? */
}

/*
 *	This routine is called after a delay, to turn the Modem Interrupts back
 *	on. The interrupts were most likely turned off because of a noisy
 *	modem interrupt, and we solve this by switching to time polling.
 *
 *	An alternative was suggested to poll the modem status, after
 *	investigation there was no perceptable difference and in a thought
 *	experiment, the overhead is far less to use Interrupts. And let it
 *	fall back to this `semi-polled' action if the interrupts become
 *	too anoying for the offending ports.
 */
static void
MuxModemExecutive (dev)
	int		dev;
{	struct tty_sc *	sp;
	int		port, s, c, instance;
	struct tty *	tp ;

	if ((dev >= 0) && (port = (sp = &sc[MUX_UNIT ((dev_t)dev)])->sc_addr)
	 && (sp->sc_tty.t_state & (TS_ISOPEN | TS_WOPEN))) {
		for (instance = 0; (instance < num_instance)
		 && (sp < first_sc[instance]); ++instance);
		s = splmux ();
		outb (port + com_cfcr, inb (port + com_cfcr) & ~CFCR_DLAB);
		outb (port + com_ier, inb (port + com_ier) | IER_EMSC);
		sp->sc_mcount = 0;
		tp = &(sc->sc_tty) ;
		/*
		 *	Simulate a modem change interrupt
		 */
		if ((c = inb (port + com_msr)) & MSR_DDCD) {
		    if (c & MSR_DCD) {
			(void)(*linesw[(int)tp->t_line].l_modem)(tp, 1);
			/* Lets not trust Mach ... */
			wakeup((caddr_t) &tp->t_wopen);
		    }
		    /*
		     * 	I would like to see local lines ignore this transition,
		     * but if the line is truly local, then the lines themselves
		     * will not make the transition. By ignoring it, at least
		     * when a modem call drops a carrier, I get a signal ... To
		     * solve a possible need (I use it personally to shut down
		     * the uucico immediately if an unexpected modem hangup
		     * occurs), I have created a different class of serial
		     * ports (cuda & cudfa) that open ignoring DCD, but will
		     * detect a drop in DCD. This later special port works well
		     * with the base NeXT uucico, and with mgetty+sendfax V0.22.
		     */
		    if (MUX_LOCAL(tp->t_dev) && !MUX_HANGUP(tp->t_dev))
			tp->t_state |= TS_CARR_ON;
		    else if ((*linesw[(int)tp->t_line].l_modem)(tp, 0) == 0)
			outb (port + com_mcr, sp->sc_mcr &= ~MCR_DTR);
		}
		if ((c & MSR_DCTS) && (tp->t_state & TS_ISOPEN)) {
		    /*
		     *	the line is up and we want to do rts/cts flow control
		     */
		    if ((MUX_FLOW(tp->t_dev) == 0) || (c & MSR_CTS)) {
			extern void	ttstart ();

			tp->t_state &=~ TS_TTSTOP;
			ttstart(tp);
		    } else
			tp->t_state |= TS_TTSTOP;
		}
		splx (s);
		IOLog("%s[%x]: noisy modem\n", MuxDeviceName, sp->sc_tty.t_dev);
	}
}

/*
 *	identity, and state are passed to us *only* if we are using the
 *	mach signalling environment. Since I have a desire to keep overhead
 *	to a minimum, we have not, and the identity variable is truly loaded
 *	with the argument (which we ignore).
 */

static void
#ifdef IF_IODevice_Worked
Muxintr (void *identity, void *state, unsigned int instance)
#else
Muxintr (unsigned int instance)
#endif
{	u_char		code;
	int		c, port, unit = 0, keep_going = 0;
	struct tty *	tp;
	struct tty_sc *	sp, *fp = first_sc[instance];

again:	sp = &fp[unit];
	tp = &sp->sc_tty;
	/*
	 *	An ongoing debate. I have been torn between actually using the
	 * Multiplex status register to determine the ports to service or to
	 * let it run free and check *all* the ports. If I use the register to
	 * drive this decision, I gain in not needing to do as much servicing.
	 * If I don't, then I ensure that all characters in all the receiver
	 * and transmitter FIFOs are dealt with and synchronized (assuming they
	 * all are running at the same high rate) giving me a lower interrupt
	 * occurrence rate amoungst the multiplexed ports.
	 *
	 *	I have decided for lower interrupt hits, but will compromise
	 * and only service the ports that are either interrupting me, or the
	 * ones that are open at the moment. I feel this compromise should
	 * offer the highest level of performance but have noticed only a
	 * slight improvement.
	 */
#	ifndef NO_MUX
		if ((sp->sc_mux != 0)	/* Not multiplexed */
		 && (sp->sc_bit != -1) && ((c = inb (sp->sc_mux)) != sp->sc_bit)
		 && (((tp->t_state & (TS_ISOPEN | TS_WOPEN)) == 0)
		  || ((sp->sc_nty->t_pflags & TP_CREAD) == 0)))
			goto another_port;
#	endif
	port = sp->sc_addr;

	/*
	 *	This is the other half of the synchronizing code. We do it here
	 * for two reasons. One, if we placed it at the no pending interrupt
	 * check, then at higher baud rates, we may lose characters. Another
	 * reason for placing it here is actually to reduce the initial
	 * overhead of picking up characters from the receiver.
	 */
	/* schedule to refill the transmit buffer if need be */
	if ((sp->sc_size) && (tp->t_outq.c_cc) && notMuxTxExecutive())
		ns_timeout ((func)MuxTxExecutive, (void *)-1L, (ns_time_t)1ULL,
		    CALLOUT_PRI_SOFTINT0);
	if (code = (inb (port + com_lsr) & LSR_RCV_MASK))
		goto receive;
	do switch ((code = inb (port + com_iir)) & IIR_IMASK) {
	case IIR_RLS:
		code = inb (port + com_lsr);
		goto receive;

	case IIR_RXTOUT:
	case IIR_RXRDY:
		code = 0;

	receive:
	    if (sp->sc_size) {
		short *	rp = sp->sc_rxin;

		/*
		 *	We want the highest speed possible to placing
		 *	the received data into the buffer.
		 */
		keep_going = 0;	/* used as a byte counter here */
		if (sp->sc_rxbuf) {
			do {	*((char *)rp) = inb (port + com_data);
				*(((char *)rp) + 1) = code;
				if (++rp >= TOP_RXBUF (sp))
					rp = sp->sc_rxbuf;
				keep_going++;
			} while (code = (inb (port + com_lsr) & LSR_RCV_MASK));
			if ((sp->sc_rxnum += keep_going) > RTS_HIWAT(sp)) {
				tp->t_state |= TS_INPUTFULL;
				outb (port + com_mcr, sp->sc_mcr &= ~MCR_RTS);
			}
		} else {do {
				c = inb (port + com_data);
			} while (code = (inb (port + com_lsr) & LSR_RCV_MASK));
		}
		sp->sc_rxin = rp;
		/*
		 *	We use the timeout executive mechanism for two reasons.
		 * This places the MuxRxExecutive at the lower priority kernel
		 * processing levels, and with the inherent delay in processing
		 * a timeout function, we also allow a buildup of processing to
		 * improve the efficiency. We use the notMuxRxExecutive() to
		 * check if we already have some processing in the queue.
		 */
		if (notMuxRxExecutive())
		    ns_timeout ((func)MuxRxExecutive, (void *)-1L,
			(ns_time_t)1ULL, CALLOUT_PRI_SOFTINT0);
	    } else {
		int	parenb = (
		((sp->sc_nty->t_pflags
		 & (TP_PARENB|TP_IGNPAR)) == (TP_PARENB)) &&
		  (((tp->t_flags & ANYP) == EVENP)
		  || (tp->t_flags & ANYP) == ODDP));
		keep_going = 0;

		do {	c = inb (port + com_data);
			if (code &~ LSR_RCV_MASK) {
				if (code & (LSR_BI | LSR_FE))
					c |= TTY_FE;
				else if ((code & LSR_PE) && parenb)
					c |= TTY_PE;
				/* if (code & LSR_OE)
				 *	keep_going = 1;
				 */
			}
			if (tp->t_state & TS_ISOPEN)
				(*linesw[(int)tp->t_line].l_rint)(c,tp);
		} while (code = (inb (port + com_lsr) & LSR_RCV_MASK));
		/* if (keep_going)
		 *	log(LOG_WARNING, "%s[%d:%d]: silo overflow\n",
		 *	    MuxDeviceName, instance, unit);
		 */
  		if ((tp->t_rawq.c_cc + tp->t_canq.c_cc) > TTHIWAT (tp)) {
			tp->t_state |= TS_INPUTFULL;
			outb (port + com_mcr, sp->sc_mcr &= ~MCR_RTS);
		}
	    }
	    keep_going = 1;
	    if ((sp->sc_flags & SC_HASFIFO)
	     || ((inb (port + com_lsr) & LSR_TSRE) == 0)
	     || ((tp->t_outq.c_cc == 0)
	      && (sp->sc_txin == sp->sc_txout)))
		break;
	    /* bug in 16450 may clear pending tx-interrupt */
	    /* FALL THROUGH */

	transmit: case IIR_TXRDY:
	    keep_going = 1;
	    tp->t_state &= ~(TS_BUSY | TS_OUTPUTBUSY | TS_FLUSH);
	    (*linesw[(int)tp->t_line].l_start) (tp);
	    break;

	default:
	    if (code & IIR_NOPEND) {
		/*
		 *	Done with this line.  If there are more lines on this
		 * IRQ, move to the next one. If we have multiple lines on this
		 * IRQ and at least one had something, start over until we get
		 * a clean sweep to avoid losing an interrupt.
		 */
		c = 0xFF;
#		ifndef NO_MUX
			if (sp->sc_mux)
				c = inb (sp->sc_mux);
another_port:
#		endif
		if (++unit >= num_devices[instance]) {
		    if ((keep_going || ((c & 0xFF) != 0xFF))
		     && (num_devices[instance] > 1)) {
			keep_going = 0;
			unit = 0;
		    } else {
#			ifdef PASS_SendInterrupt
				/*
				 *	Lets *not* pass the interrupt signal up
				 * to the Objects since there should be
				 * *nothing* they should do with it ...
				 */
				IOSendInterrupt (identity, state,
				    IO_DEVICE_INTERRUPT_MSG);
#			endif
			return;
		    }
		}
		goto again;
	    }

	    /* log is *not* possible at interrupt priorities
	     * log(LOG_WARNING,"%s[%d:%d]: weird interrupt, id 0x%x\n",
	     *	    MuxDeviceName, instance, unit, code);
	     */
	    if ((sp->sc_flags & SC_HASFIFO) == 0)
		goto transmit;	/* maybe it was a corrupted tx-interrupt */
	    /* FALL THROUGH (an indirect way of counting bad interrupts too) */

	case IIR_MLSC:
	    /*
	     *	If we get too many interrupts here, a problem with the 16450
	     * chip, and *any* port with a floating or noisy set of Modem
	     * control lines, then we will have to shut down listening to
	     * the lines until a suitable delay (500ms should do). It was
	     * determined experimentally that there is a wide margin between
	     * a floating or malfunctioning port and a properly operating port.
	     * A value of 8 modem interrupts is half way between these two
	     * extremes.
	     */
	    if (++(sc->sc_mcount) == 8) {
		outb (port + com_cfcr, inb (port + com_cfcr) & ~CFCR_DLAB);
		outb (port + com_ier, inb (port + com_ier) & ~IER_EMSC);
		ns_timeout ((func)MuxModemExecutive, (void *)(long)(tp->t_dev),
		    (ns_time_t)500000000ULL, CALLOUT_PRI_SOFTINT0);
	    }
	    else if (sc->sc_mcount > 100)
		sc->sc_mcount = 100;
	    keep_going = 1;
	    if ((c = inb (port + com_msr)) & MSR_DDCD) {
		/*
		 *	This should send a wakeup call. There is some protection
		 * for noisy modem lines as noted above.
		 */
		if (c & MSR_DCD) {
			(void)(*linesw[(int)tp->t_line].l_modem)(tp, 1);
			/* Lets not trust Mach ... */
			wakeup((caddr_t) &tp->t_wopen);
		}
		/*
		 * 	I would like to see local lines ignore this transition,
		 * but if the line is truly local, then the lines themselves
		 * will not make the transition. By ignoring it, at least when
		 * a modem call drops a carrier, I get a signal ... To solve a
		 * possible need (I use it personally to shut down the uucico
		 * immediately if an unexpected modem hangup occurs), I have
		 * created a different class of serial ports (cuda & cudfa)
		 * that open ignoring DCD, but will detect a drop in DCD.
		 */
		else if (MUX_LOCAL(tp->t_dev) && !MUX_HANGUP(tp->t_dev))
			tp->t_state |= TS_CARR_ON;
		else if ((*linesw[(int)tp->t_line].l_modem)(tp, 0) == 0)
			outb (port + com_mcr, sp->sc_mcr &= ~MCR_DTR);
	    }
	    if ((c & MSR_DCTS) && (tp->t_state & TS_ISOPEN)) {
		/*
		 *	the line is up and we want to do rts/cts flow control
		 */
		if ((MUX_FLOW(tp->t_dev) == 0) || (c & MSR_CTS)) {
			extern void	ttstart ();

			tp->t_state &=~ TS_TTSTOP;
			ttstart(tp);
		} else	tp->t_state |= TS_TTSTOP;
	    }
	    break;
	} while (1);
	/* NOTREACHED */
}

static int
tiocm2mcr (data)
	int	data;
{	int	m = 0;

	if (data & TIOCM_DTR)
		m |= MCR_DTR;
	if (data & TIOCM_RTS)
		m |= MCR_RTS;
	return (m);
}

/*
 *	Replacing kernel routine with one that does not panic.
 *	Guilty Party: Mark Salyzyn  mark@ve6mgs.ampr.ab.ca
 */
static struct posix_proc *
get_posix_proc (pg_id)
	pid_t			pg_id ;
{	struct posix_proc *	p ;

	for (p = posix_proc_hash[pg_id & 0x3F];
	     p && (p->p_pid != pg_id);
	     p = p->p_next_posix_proc) ;
	return (p) ;
}

static int
Muxioctl (dev, cmd, data, flag)
	int		cmd, flag;
	dev_t		dev;
	caddr_t		data;
{	int		unit = MUX_UNIT (dev);
	struct tty_sc *	sp = &sc[unit];
	struct tty *	tp = &sp->sc_tty;
	int		error, port;
	extern int	ttioctl ();

#	if (defined(DEBUG))
		IOLog ("%s: Muxioctl (%x, %x, ...) ", MuxDeviceName, dev, cmd);
#	endif
	if ((error = (*linesw[(int)tp->t_line].l_ioctl)(tp,cmd,data,flag)) >= 0)
		return (error);
	/* Flush output queue completely for commands that ask */
	if ((sp->sc_size) && ((cmd == TIOCSETP) || (cmd == TIOCDRAIN)
	 || (cmd == TIOCSETAW) || (cmd == TIOCSETAF)
	 || ((cmd == TIOCFLUSH) && (flag == FWRITE)))) {
		/*
		 *	Reuse of variables is performed as a code optimization
		 * that may add confusion to the code. However, the code uses
		 * up less stack space if the multitude of variables is kept
		 * to a minimum. Reuse error as a loop counter here.
		 */
		error = 300;	/* Wait about 3 seconds maximum */
		if ((port = sp->sc_addr) == 0)
			return (ENOTTY);
		/* This will force a delay at least until characters are gone.*/
		while ((tp->t_outq.c_cc || (sp->sc_txin != sp->sc_txout)
		  || ((inb (port+com_lsr) & LSR_TSRE) == 0))
		 && (--error > 0)) {
			tp->t_state &= ~(TS_BUSY | TS_OUTPUTBUSY | TS_FLUSH);
			(*linesw[(int)tp->t_line].l_start) (tp);
			IOSleep (10);
		}
		if (error <= 0) {
			IOLog ("%s: %x giving up on flush: ",
			    MuxDeviceName, dev);
			if (tp->t_outq.c_cc) {
				IOLog ("still characters in outq");
				error = 0;
			}
			if (sp->sc_txin != sp->sc_txout) {
				IOLog ("%stransmit-buffer",
				    error ? "still characters in "
					  : ", ");
				error = 0;
			}
			if ((inb (port + com_lsr) & LSR_TSRE) == 0)
				IOLog ("%sUART",
				    error ? "still characters in "
					  : ", ");
			IOLog ("\n");
		}
		/*
		 *	I have opted *not* to flush the transmit double buffer
		 * after the above timeout. I hope this does not become an
		 * issue with creating garbled characters, but I figured this
		 * was an appropriate action due to my idea of timing out after
		 * 3 seconds ...
		sp->sc_txout = sp->sc_txin;
		 */
	}
	if ((cmd == TIOCSETP) || (cmd == TIOCDRAIN) || (cmd == TIOCSETAW)
	 || (cmd == TIOCSETAF) || ((cmd == TIOCFLUSH) && (flag == FREAD))) {
		sp->sc_rxout = sp->sc_rxin;
		sp->sc_rxnum = 0;
	}
#if (defined(POSIX_KERN))
	switch (cmd) {
	/*
	 * Set terminal process group.
	 */
	case TIOCSPGRP: {
		extern struct pgrp *	pgfind ();
		struct pgrp *		pgrp = pgfind(*(int *)data);
		struct posix_proc *	px;
		struct proc *		p;
#		if NeXT
			struct utask *up ;
#		endif

		p = getproc () ;
#		if (defined(DEBUG))
			IOLog ("pgrp(0x%lx) = {\n\t0x%lx, 0x%lx, 0x%lx,\n\t%d, %d\n};\n",
			    (long int)pgrp, (long int)(pgrp->pg_hforw),
			    (long int)(pgrp->pg_mem),
			    (long int)(pgrp->pg_session), pgrp->pg_id,
			    pgrp->pg_jobc) ;
			IOLog ("proc(0x%lx) = {\n\t0x%lx, 0x%lx, 0x%lx, 0x%lx,\n\t%d, %d, %d, %d, %d, %d, %d, %d, %d,\n\t0x%x, 0x%x, 0x%x, 0x%x, 0x%x,\n\t%d, %d, %d, %d ...\n};\n",
			    (long int)p, (long int)(p->p_link),
			    (long int)(p->p_rlink), (long int)p->p_nxt,
			    (long int)p->p_prev, p->p_usrpri, p->p_pri,
			    p->p_cpu, p->p_stat, p->p_time, p->p_nice,
			    p->p_debugger, p->p_posix, p->p_cursig, p->p_sig,
			    p->p_sigmask, p->p_sigignore, p->p_sigcatch,
			    p->p_flag, p->p_uid, p->p_pgrp, p->p_pid,
			    p->p_ppid);
			IOLog ("posix_proc_hash(0x%lx) = {\n\t",
			    (long int)posix_proc_hash);
			{	struct posix_proc **pp;
				int	len = ((29+7)>>3)<<3 ;
				char	buffer[16] ;

				pp = posix_proc_hash ;
				do {	if (pp != posix_proc_hash) {
						IOLog (", ") ;
						len += 2 ;
					}
					if (len > 70) {
						IOLog ("\n\t") ;
						len = ((29+7)>>3)<<3 ;
					}
					IOLog ("{") ;
					len += 1 ;
					px = *pp ;
					while (px) {
						if (*pp != px) {
							IOLog (",") ;
							++len ;
						}
						sprintf (buffer,"%d",px->p_pid);
						IOLog (buffer) ;
						len += strlen (buffer) ;
						px = px->p_next_posix_proc ;
					}
					IOLog ("}");
					++len ;
				} while ((++pp - posix_proc_hash) < 64) ;
				IOLog ("}\n") ;
			}
#		endif
		if (!(px = get_posix_proc (pgrp->pg_id)))
			break ; /* return (ENOTTY); */
#		if (defined(DEBUG))
			IOLog ("%s[%x]: TIOCSPGRP %sISCTTY, %sSESSION\n",
			    MuxDeviceName, dev,
			    (isctty(p, px, sp->sc_nty)) ? "" : "!",
			    (pgrp == NULL) ? "!PGRP, " :
				(pgrp->pg_session == px->p_session)
				    ? "" : "!");
			IOLog ("%s[%x]: isctty p_session=%lx t_session=%lx\n",
			    MuxDeviceName, dev, (long int)px->p_session,
			    (long int)sp->sc_nty->t_session) ;
			IOLog ("%s[%x]: pg_session=%lx p_flags=%x\n",
			    MuxDeviceName, dev, (long int)pgrp->pg_session,
			    p->p_flag) ;
#		endif
		/*
		 *	if (!isctty(p, px, sp->sc_nty))
		 *		return (ENOTTY);
		 */
		if (pgrp == NULL || pgrp->pg_session != px->p_session)
			break ; /* return (EPERM); */
#		if (defined(DEBUG))
			IOLog ("sp->sc_nty->t_posix_pgrp=0x%x\n",
			    (int)sp->sc_nty->t_posix_pgrp) ;
			IOLog ("pgrp=0x%x, tp->t_pgrp=%d, pgrp->pg_id=%d\n",
			    (int)pgrp, tp->t_pgrp, pgrp->pg_id) ;
#		endif
		px->p_posix_pgrp = sp->sc_nty->t_posix_pgrp = pgrp;
		p->p_pgrp = tp->t_pgrp = pgrp->pg_id;
		sp->sc_nty->t_session = pgrp->pg_session ;
#		if NeXT
			if (up = u_task ())
				up->uu_ttyp = tp ;
#		else
			p->p_ttyp =
#		endif
		pgrp->pg_session->s_ttyp = tp ;
#		if NeXT
			if (up)
				up->uu_ttyd = dev ;
#		else
			p->p_ttyd =
#		endif
		pgrp->pg_session->s_ttyd = dev ;
		p->p_flag |= SCTTY ;
		(void)ttioctl(tp, cmd, data, flag) ;
		return (0) ; /* break; */
	}

	case TIOCGPGRP: {
		struct posix_proc *	px;
		struct proc *		p;

#		if (defined(DEBUG))
			IOLog ("%s[%x]: TIOCGPGRP ", MuxDeviceName, dev);
#		endif
		p = getproc ();
		if (!(px = get_posix_proc (tp->t_pgrp)))
			break ; /* return (ENOTTY); */
#		if (defined(DEBUG))
			IOLog ("%s[%x]: isctty p_session=%lx t_session=%lx\n",
			    MuxDeviceName, dev, (long int)px->p_session,
			    (long int)sp->sc_nty->t_session) ;
			IOLog ("%s[%x]: p_flags=%x\n",
			    MuxDeviceName, dev, p->p_flag) ;
#		endif
		if (!isctty(p, px, sp->sc_nty)) {
#			if (defined(DEBUG))
				IOLog ("!ISCTTY\n");
#			endif
			break ; /* return (ENOTTY); */
		}
#		if (defined(DEBUG))
			if (px->p_posix_pgrp)
				IOLog ("PGRP_ID=%d\n",
				    px->p_posix_pgrp->pg_id);
			else	IOLog ("!PGRP\n");
#		endif
		*(int *)data = sp->sc_nty->t_posix_pgrp
		    ? sp->sc_nty->t_posix_pgrp->pg_id : -1;
		return (0) ; /* break; */
	}
	}
#endif
	if ((error = ttioctl(tp, cmd, data, flag)) >= 0) {
		if (cmd == TIOCSETP  || cmd == TIOCSETN || cmd == TIOCLBIS ||
		    cmd == TIOCLBIC  || cmd == TIOCLSET || cmd == TIOCSETA ||
		    cmd == TIOCSETAW || cmd == TIOCSETAF)
			Muxparam(unit);
#		if (defined(DEBUG))
			IOLog (" ttioctl\n");
#		endif
		return (error);
	}
	if ((port = sp->sc_addr) == 0)
		return (ENOTTY);
	switch (cmd) {
	case TIOCSBRK:
		outb (port + com_cfcr, inb (port + com_cfcr) | CFCR_SBREAK);
		break;

	case TIOCCBRK:
		outb (port + com_cfcr, inb (port + com_cfcr) & ~CFCR_SBREAK);
		break;

	case TIOCSDTR:
		/* only DTR, RTS exist */
		outb (port + com_mcr, sp->sc_mcr |= MCR_DTR );
		break;

	case TIOCCDTR:
		/* only DTR, RTS exist */
		outb (port + com_mcr, sp->sc_mcr &= ~MCR_DTR);
		break;

	case TIOCMSET:
		outb (port + com_mcr, sp->sc_mcr = (sp->sc_mcr & ~MCR_ON)
			| tiocm2mcr (*(int *)data));
		break;

	case TIOCMBIS:
		outb (port + com_mcr, sp->sc_mcr |= tiocm2mcr (*(int *)data));
		break;

	case TIOCMBIC:
		outb (port + com_mcr, sp->sc_mcr &= ~tiocm2mcr (*(int *)data));
		break;

	case TIOCMGET:
		/*
		 *	Reuse of variables is performed as a code optimization
		 * that may add confusion to the code. However, the code uses
		 * up less stack space if the multitude of variables is kept
		 * to a minimum. Reuse unit and error here, please do not get
		 * confused ...
		 */
		unit = sp->sc_mcr /* inb (port + com_mcr) */ ;
		error = TIOCM_LE;

		if (unit & MCR_DTR)
			error |= TIOCM_DTR;
		if (unit & MCR_RTS)
			error |= TIOCM_RTS;
		unit = inb (port + com_msr);
		if (unit & MSR_CTS)
			error |= TIOCM_CTS;
		if (unit & MSR_DCD)
			error |= TIOCM_CAR;
		if (unit & MSR_DSR)
			error |= TIOCM_DSR;
		if (unit & (MSR_RI | MSR_TERI))
			error |= TIOCM_RI;
		*(int *)data = error;
		break;
	default:
#		if (defined(DEBUG))
			IOLog (" Invalid command\n") ;
#		endif
		return (ENOTTY);
	}
#	if (defined(DEBUG))
		IOLog (" Ok return\n") ;
#	endif
	return (0);
}

/*
 * Stop output on a line.
 */
static void
/*ARGSUSED*/
Muxstop (tp, flag)
	struct tty *	tp;
	int		flag;
{	struct tty_sc *	sp = &sc[MUX_UNIT(tp->t_dev)];
	/*
	 *	Mach sez we don't need the MUX_IPL level change here. I say
	 *	since this operation is not atomic I will keep it here
	 *	anyways.
	 */
	int		s = splmux ();

	if ((tp->t_state & (TS_BUSY | TS_OUTPUTBUSY))
	 && ((tp->t_state & TS_TTSTOP) == 0))
		tp->t_state |= TS_FLUSH;
	/*
	 *	VINTR will cause a call to us with FREAD and FWRITE set if the
	 * flush operation is enabled on interrupt. This should flush the
	 * character queue in our double buffer.
	 */
	if (flag & FREAD) {
		sp->sc_rxout = sp->sc_rxin;
		sp->sc_rxnum = 0;
	}
	if (flag & FWRITE)
		sp->sc_txout = sp->sc_txin;
	splx (s);
}

@end

#endif
